from __future__ import annotations

import numpy as np
import pytest

import skretrieval as skr
from skretrieval.retrieval.observation import SimulatedNadirObservation
from skretrieval.util import configure_log


@pytest.mark.parametrize("minimizer", ["rodgers", "scipy"])
def test_simulated_retrieval(minimizer: str):
    configure_log()
    obs = SimulatedNadirObservation(
        cos_sza=0.6,
        cos_viewing_zenith=1.0,
        reference_latitude=20,
        reference_longitude=0,
        sample_wavelengths=np.arange(280, 350, 0.5),
        state_adjustment_factors={
            "o3": {"vmr": {"scale": 2}},
            "lambertian_albedo": {"albedo": {"scale": 0.5}},
        },
    )

    ret = skr.Retrieval(
        obs,
        minimizer=minimizer,
        state_kwargs={
            "altitude_grid": np.arange(0, 70000, 1000),
            "absorbers": {
                "o3": {
                    "prior_influence": 5e0,
                    "tikh_factor": 1e-2,
                    "log_space": False,
                    "min_value": 0,
                    "max_value": 1,
                    "prior": {"type": "mipas"},
                },
            },
            "surface": {
                "lambertian_albedo": {
                    "prior_influence": 0,
                    "tikh_factor": 1e-2,
                    "log_space": False,
                    "wavelengths": np.array([280, 360]),
                    "initial_value": 0.5,
                }
            },
        },
    )

    _ = ret.retrieve()



from __future__ import annotations

import numpy as np
import xarray as xr
from sasktran import Geometry, LineOfSight

import skretrieval.core.radianceformat as radianceformat
from skretrieval.core import OpticalGeometry
from skretrieval.core.lineshape import LineShape
from skretrieval.legacy.core.sensor import Sensor


class Pixel(Sensor):
    """
    A single pixel with corresponding horizontal field of view, vertical field of view, and wavelength line shape.
    """

    def __init__(
        self,
        mean_wavel_nm: float,
        pixel_shape: LineShape,
        vert_fov: LineShape,
        horiz_fov: LineShape,
    ):
        """
        Parameters
        ----------
        mean_wavel_nm : float
            Mean wavelength of the pixel
        pixel_shape: LineShape
            Lineshape for the wavelength
        vert_fov: LineShape
            Vertical field of view
        horiz_fov: LineShape
            Horizontal field of view
        """
        self._wavelength_nm = mean_wavel_nm
        self._pixel_shape = pixel_shape
        self._vert_fov = vert_fov
        self._horiz_fov = horiz_fov

    def measurement_geometry(self, optical_geometry: OpticalGeometry):
        return [
            LineOfSight(
                optical_geometry.mjd,
                optical_geometry.observer,
                optical_geometry.look_vector,
            )
        ]

    def model_radiance(
        self,
        optical_geometry: OpticalGeometry,
        model_wavel_nm: np.array,
        model_geometry: Geometry,
        radiance: np.array,
        wf=None,
    ):
        wavel_interp, los_interp = self._construct_interpolation(
            hires_wavel_nm=model_wavel_nm,
            model_geometry=model_geometry,
            optical_axis=optical_geometry,
        )

        modelled_radiance = np.einsum("ij,jk...,kl", wavel_interp, radiance, los_interp)

        meas_los = self.measurement_geometry(optical_geometry)[0]

        data = xr.Dataset(
            {
                "radiance": (["wavelength", "los"], modelled_radiance),
                "mjd": (["los"], [meas_los.mjd]),
                "los_vectors": (
                    ["los", "xyz"],
                    np.array(meas_los.look_vector).reshape((1, 3)),
                ),
                "observer_position": (
                    ["los", "xyz"],
                    np.array(meas_los.observer).reshape((1, 3)),
                ),
            },
            coords={
                "wavelength": [self.measurement_wavelengths()],
                "xyz": ["x", "y", "z"],
            },
        )

        if wf is not None:
            modelled_wf = np.einsum("ij,jkl,k...", wavel_interp, wf, los_interp)

            data["wf"] = (["wavelength", "los", "perturbation"], modelled_wf)

        return radianceformat.RadianceGridded(data)

    def required_wavelengths(self, res_nm: float) -> np.array:
        """
        Recommended wavelengths to use in the high resolution input radiance

        Parameters
        ----------
        res_nm: float
            Desired model resolution

        Returns
        -------
        np.array
            Array of wavelengths
        """
        bounds = self._pixel_shape.bounds()
        return np.arange(bounds[0], bounds[1], res_nm) + self._wavelength_nm

    def measurement_wavelengths(self):
        """
        Central wavelength of the pixel
        """
        return self._wavelength_nm

    @staticmethod
    def radiance_format():
        return radianceformat.RadianceGridded

    def _construct_wavelength_interpolator(self, hires_wavel_nm):
        """
        Internally constructs the matrix used for the wavelength interpolation
        """
        wavel_interpolator = np.zeros((1, len(hires_wavel_nm)))

        wavel_interpolator[0, :] = self._pixel_shape.integration_weights(
            self._wavelength_nm, hires_wavel_nm
        )

        return wavel_interpolator

    def _construct_los_interpolator(
        self, model_geometry, optical_axis: OpticalGeometry
    ):
        """
        Internally constructs the matrix used for the line of sight interpolation
        """
        x_axis = np.array(optical_axis.look_vector)
        vert_normal = np.cross(np.array(x_axis), np.array(optical_axis.local_up))
        vert_normal = vert_normal / np.linalg.norm(vert_normal)
        vert_y_axis = np.cross(vert_normal, x_axis)

        horiz_y_axis = vert_normal

        horiz_angle = []
        vert_angle = []

        for los in model_geometry.lines_of_sight:
            vert_angle.append(
                np.arctan2(
                    np.dot(los.look_vector, vert_y_axis),
                    np.dot(los.look_vector, x_axis),
                )
            )
            horiz_angle.append(
                np.arctan2(
                    np.dot(los.look_vector, horiz_y_axis),
                    np.dot(los.look_vector, x_axis),
                )
            )

        horiz_interpolator = self._horiz_fov.integration_weights(
            0, np.array(horiz_angle)
        )
        vert_interpolator = self._vert_fov.integration_weights(0, np.array(vert_angle))

        # TODO: Only valid for exponential distributions?
        los_interpolator = np.zeros((len(vert_interpolator), 1))
        los_interpolator[:, 0] = horiz_interpolator * vert_interpolator
        los_interpolator[:, 0] /= np.nansum(los_interpolator)

        return los_interpolator

    def _construct_interpolation(
        self, hires_wavel_nm, model_geometry, optical_axis: OpticalGeometry
    ):
        """
        Creates both the wavelength and line of sight interpolators.  This is done so we can write

        output_radiance = (wavel_interp) @ (input_radiance) @ (los_interp)
        """
        # Radiance is [wavel, los], which gets collapsed down to a single measurement

        wavel_interpolator = self._construct_wavelength_interpolator(hires_wavel_nm)
        los_interpolator = self._construct_los_interpolator(
            model_geometry, optical_axis
        )

        # Collapsed radiance is is wavel_interpolator @ radiance @ los_interpolator
        return wavel_interpolator, los_interpolator



from __future__ import annotations

import abc
import dataclasses
from copy import copy

import numpy as np

from skretrieval.retrieval.statevector import StateVectorElement
from skretrieval.retrieval.tikhonov import (
    two_dim_vertical_first_deriv,
    two_dim_vertical_second_deriv,
)


@dataclasses.dataclass
class Prior:
    """
    A dataceass to hold a prior state of size (n,) and an inverse covariance of size (n, n)
    """

    state: np.array
    inverse_covariance: np.ndarray


class BasePrior:
    """
    Base class to handle prior states
    """

    @property
    @abc.abstractmethod
    def state(self) -> np.array:
        """
        The prior state $x_a$ of size (n,)
        """

    @property
    @abc.abstractmethod
    def inverse_covariance(self):
        """
        The inverse covariance of the prior state $S_a^{-1}$ of size (n, n)
        """

    def __mul__(self, other):
        return MultipliedPrior(self, other)

    __rmul__ = __mul__

    def __add__(self, other):
        return AdditivePrior(self, other)

    def init(self, sv: StateVectorElement, sv_slice: slice | None = None):
        pass


class MultipliedPrior(BasePrior):
    def __init__(self, prior: BasePrior, multiplier: float):
        """
        A prior where the inverse covariance is multiplied by a scalar, the prior state remains unchanged

        Parameters
        ----------
        prior : BasePrior
        multiplier : float
        """
        self._prior = prior
        self._multiplier = multiplier

    @property
    def state(self):
        return self._prior.state

    @property
    def inverse_covariance(self):
        return self._prior.inverse_covariance * self._multiplier

    def init(self, sv: StateVectorElement, sv_slice: slice | None = None):
        self._prior.init(sv, sv_slice)


class AdditivePrior(BasePrior):
    def __init__(self, prior1: BasePrior, prior2: BasePrior):
        """
        A prior where two priors are added together.  This results in a sum of the inverse covariance,
        and then a new prior state x_a

        Parameters
        ----------
        prior1 : BasePrior
        prior2 : BasePrior
        """
        self._prior1 = prior1
        self._prior2 = prior2

    @property
    def state(self):
        # Have to solve the system to get the equivalent prior state

        inv_S_a_1 = self._prior1.inverse_covariance
        inv_S_a_2 = self._prior2.inverse_covariance
        x_a_1 = self._prior1.state
        x_a_2 = self._prior2.state

        full_inv_S_a = inv_S_a_1 + inv_S_a_2

        rhs = inv_S_a_1 @ x_a_1 + inv_S_a_2 @ x_a_2

        # For some priors the inverse covariance will be singular
        try:
            return np.linalg.solve(full_inv_S_a, rhs)
        except np.linalg.LinAlgError:
            # If the inverse covariance is singular, we can't solve the system
            # TODO: Is this actually right? It seems okay in most cases, but in general
            # i'm not so sure
            return 0.5 * (x_a_1 + x_a_2)

    @property
    def inverse_covariance(self):
        return self._prior1.inverse_covariance + self._prior2.inverse_covariance

    def init(self, sv: StateVectorElement, sv_slice: slice | None = None):
        self._prior1.init(sv, sv_slice)
        self._prior2.init(sv, sv_slice)


class VerticalPrior(BasePrior):
    def __init__(self, altitudes: np.array):
        self._altitudes = altitudes


class VerticalTikhonov(VerticalPrior):
    def __init__(
        self,
        order: int,
        prior_state: np.array = None,
        tikhonov: np.array = None,
    ):
        """
        A prior that is constructed as a Tikhonov constraint.

        Parameters
        ----------
        order : int
            Order of the Tikhonov constraint, only 1 and 2 are supported
        prior_state : np.array, optional
            Prior state. If set to None a zero prior is used, by default None
        tikhonov : np.array, optional
            Array of factors to multiply the constraint by, by default None
        """
        self._tikhonov = tikhonov
        self._prior_state = prior_state
        self._order = order

    def init(self, sv: StateVectorElement, sv_slice: slice | None = None):
        n = len(sv.state()[sv_slice])

        if self._order == 1:
            self._gamma = two_dim_vertical_first_deriv(1, n, factor=1)
        elif self._order == 2:
            self._gamma = two_dim_vertical_second_deriv(1, n, factor=1)
        else:
            msg = f"Order {self._order} not supported."
            raise ValueError(msg)

        if self._tikhonov is not None:
            # Scale by the weights
            self._gamma *= self._tikhonov[np.newaxis, :]

        self._prior = Prior(
            inverse_covariance=self._gamma.T @ self._gamma,
            state=(np.zeros(n) if self._prior_state is None else self._prior_state),
        )

    @property
    def state(self):
        return self._prior.state

    @property
    def inverse_covariance(self):
        return self._prior.inverse_covariance


class ManualPrior(BasePrior):
    def __init__(self, state: np.array, inverse_covariance: np.array):
        """
        A prior that is manually specified, both the prior state and it's covariance

        Parameters
        ----------
        state : np.array
        inverse_covariance : np.array
        """
        self._state = state
        self._inverse_covariance = inverse_covariance

    @property
    def state(self):
        return self._state

    @property
    def inverse_covariance(self):
        return self._inverse_covariance

    def init(self, sv: StateVectorElement, sv_slice: slice | None = None):
        pass


class ConstantDiagonalPrior(BasePrior):
    def __init__(self, value: float = 1.0):
        """
        A prior that is constant along the diagonal. The initial state is pulled from
        the StateVectorElement upon initialization.

        Parameters
        ----------
        value : float, optional
            _description_, by default 1.0
        """
        self._value = value

    def init(self, sv: StateVectorElement, sv_slice: slice | None = None):
        n = len(sv.state()[sv_slice])
        self._prior = Prior(
            inverse_covariance=np.eye(n) * self._value,
            state=copy(sv.state()[sv_slice]),
        )

    @property
    def state(self):
        return self._prior.state

    @property
    def inverse_covariance(self):
        return self._prior.inverse_covariance



from __future__ import annotations

import logging
from copy import copy

import numpy as np
import sasktran as sk
import xarray as xr
from scipy import sparse

from skretrieval.legacy.tomography.grids import OrbitalPlaneGrid


class EngineHRTwoDim:
    def __init__(
        self,
        geometry: list[sk.Geometry],
        atmosphere: sk.atmosphere,
        grid: OrbitalPlaneGrid,
        wavelength: np.ndarray = None,
        max_difference_seconds=100,
        common_options: dict | None = None,
        grid_spacing: float = 1000,
    ):
        """
        Radiative transfer model using the HR engine that calculates radiances for an entire orbit.

        Radiative transfer calculations are split into chunks depending on the time delta between subsequent images.

        Parameters
        ----------
        geometry : List[sk.Geometry]
            A list of geometries for subsequent images/scans
        atmosphere : sk.Atmosphere
            The atmosphere
        grid : OrbitalPlaneGrid
            The grid defining the geometry of the orbit
        wavelength : np.ndarray
            Wavelengths to perform the calculation at
        max_difference_seconds : float, optional
            Time delta in seconds used to determine which radiative transfer calculations should be combined together.
            Default is 100s which roughly corresponds to the sun moving an amount equivalent to the width of the solar
            disk
        common_options : dict
            RTM options
        grid_spacing : float, optional
            Grid spacing of the model, default 1000 m
        """
        self._atmo = atmosphere
        self._grid = grid
        self._wavelength = wavelength
        self._geometry = geometry
        self._max_difference_seconds = max_difference_seconds
        self._grid_spacing = grid_spacing

        self._construct_segments()
        if common_options is not None:
            self._common_options = common_options
        else:
            self._common_options = {}

    def calculate_radiance(
        self, full_stokes_vector=False, stokes_orientation="geographic"
    ):
        """
        Calculates the radiance for the entire orbit.

        Data is returned as a dictionary with keys 'radiance' that contains a list of radiances for each input
        image/scan, and weighting function keys if calculated.
        """
        # Calculate the radiance separately for every geo_segment
        output = {}
        all_segment_radiances = []
        all_angleidx = []
        i = 0
        for segment in self._geo_segments:
            options = copy(self._common_options)
            new_geometry = self._combined_geometry(segment)
            geodetic = sk.Geodetic()
            geodetic.from_lat_lon_alt(
                new_geometry.reference_point[0],
                new_geometry.reference_point[1],
                new_geometry.reference_point[2],
            )

            local_angles, angleidx, normalandreference = self._grid.get_local_plane(
                geodetic.location
            )
            options["opticalanglegrid"] = np.rad2deg(local_angles)
            options["opticalnormalandreference"] = normalandreference

            engine = sk.EngineHR(new_geometry, self._atmo, self._wavelength, options)
            engine.atmosphere_dimensions = 2
            engine.grid_spacing = self._grid_spacing

            logging.debug("Calculating segment radiance %i", i)
            segment_radiance = engine.calculate_radiance(
                "xarray",
                full_stokes_vector=full_stokes_vector,
                stokes_orientation=stokes_orientation,
            )

            logging.debug("   Done")
            i += 1
            all_segment_radiances.append(segment_radiance.drop("wf_brdf"))
            all_angleidx.append(angleidx)

        logging.debug("all segments calculated")
        if self._common_options.get("calcwf", 0) == 3:
            # Calculating weighting functions and we have to convert the weighting functions to the overall grid
            num_los = [len(ds["los"]) for ds in all_segment_radiances]
            num_angles = len(self._grid._angles)

            if "wfheights" in self._common_options:
                num_alts = len(self._common_options["wfheights"])
            else:
                num_alts = 100  # Default

            var_names = np.array(list(all_segment_radiances[0].keys()))
            wf_names = var_names[[name.startswith("wf_") for name in var_names]]

            for name in wf_names:
                # Full WF matrix is going to be [num_wavel, num_los, num_angles * num_alts]
                sparse_wf = [
                    sparse.lil_matrix((np.nansum(num_los), num_angles * num_alts))
                    for w in self._wavelength
                ]

                for angleidx, segment_radiance, los_end, nl in zip(
                    all_angleidx, all_segment_radiances, np.cumsum(num_los), num_los
                ):
                    # Raw wf's are stored with altitude being the fastest varying dimension
                    pert_slice = slice(
                        angleidx[0] * num_alts, (angleidx[-1] + 1) * num_alts
                    )
                    los_slice = slice(los_end - nl, los_end)

                    for w_idx in range(len(self._wavelength)):
                        sparse_wf[w_idx][los_slice, pert_slice] = segment_radiance[
                            name
                        ].to_numpy()[w_idx, :, :]
                for w_idx, _wf in enumerate(sparse_wf):
                    sparse_wf[w_idx] = sparse_wf[w_idx].tocsc()
                output[name] = sparse_wf

                for idx, segment in enumerate(all_segment_radiances):
                    all_segment_radiances[idx] = segment.drop([name])

        concat_rad = xr.concat(all_segment_radiances, dim="los")

        image_radiances = []
        cur_idx = 0
        for geo in self._geometry:
            image_radiances.append(
                concat_rad.isel(los=slice(cur_idx, cur_idx + len(geo.lines_of_sight)))
            )
            cur_idx += len(geo.lines_of_sight)

        output["radiance"] = image_radiances

        logging.debug("Radiative transfer complete")
        return output

    def _construct_segments(self):
        # calculate the mean mjds of every geometry object
        mjds = []
        for geo in self._geometry:
            geo_mjd = [los.mjd for los in geo.lines_of_sight]
            mjds.append(np.nanmean(geo_mjd))

        mjds = np.asarray(mjds)

        geo_segment_idx = []
        temp_idx = [0]

        current_start_seconds = 0
        mjd_differences = np.cumsum(np.diff(mjds) * 3600 * 24)
        for idx, difference in enumerate(mjd_differences):
            if (difference - current_start_seconds) > self._max_difference_seconds:
                current_start_seconds = difference
                geo_segment_idx.append(temp_idx)
                temp_idx = [idx + 1]
            else:
                temp_idx.append(idx + 1)
        geo_segment_idx.append(temp_idx)

        self._geo_segments = geo_segment_idx

    def _combined_geometry(self, segment):
        all_lines_of_sight = np.concatenate(
            [g.lines_of_sight for g in [self._geometry[s] for s in segment]]
        )

        new_geometry = sk.Geometry()
        new_geometry.lines_of_sight = all_lines_of_sight

        mean_loc = np.nanmean(
            np.vstack([los.tangent_location().location for los in all_lines_of_sight]),
            axis=0,
        )

        mean_mjd = np.nanmean([los.mjd for los in all_lines_of_sight])

        geod = sk.Geodetic()
        geod.from_xyz(mean_loc)

        new_reference_point = [geod.latitude, geod.longitude, 0, mean_mjd]
        new_geometry.reference_point = new_reference_point

        return new_geometry



from __future__ import annotations

import numpy as np
from scipy import sparse
from scipy.optimize import least_squares, minimize
from scipy.optimize._optimize import MemoizeJac

from skretrieval.core.radianceformat import RadianceBase
from skretrieval.retrieval import ForwardModel, Minimizer, RetrievalTarget
from skretrieval.retrieval.erroranalysis import estimate_error


class SciPyMinimizer(Minimizer):
    def __init__(
        self,
        method="trf",
        max_nfev=20,
        ftol=1e-3,
        xtol=1e-36,
        x_scale="jac",
        tr_solver="exact",
        apply_state_scaling=False,
        include_bounds=False,
        num_passes=1,
        **kwargs,
    ) -> None:
        """
        A minimization wrapper around Scipy's least_squares function

        Parameters
        ----------
        method : str, optional
            Minimization method, see scipy.least_squares, by default "trf".
            Recommended to only use "lm" or "trf".
        max_nfev : int, optional
            Maximum function evalations, see scipy.least_squares, by default 20
        ftol : _type_, optional
            Tolerance on the cost function, see sci, by default 1e-3
        xtol : _type_, optional
            Tolerance on the change in state, by default 1e-36
        x_scale : str, optional
            Internal scaling applyed by the minimizer, by default "jac"
        tr_solver : str, optional
            For the "trf" method, how to solve the trust region problem, by default "exact"
        apply_state_scaling: bool, optional
            If true, then the state vector is scaled relative to the apriori in the solver, useful
            when the state vector elements are of largely varying magnitudes and you have a well
            specified prior, by default False
        include_bounds : bool, optional
            If true, then bounds are included inside the solver. Only has an effect
            weth method is "trf", by default False
        num_passes : int, optional
            Number of passes to do through the minimizer. After each pass the noise covariance is adjusted
            where measurements with large residuals are given less weight on the next pass, by default 1
        """
        self._method = method
        self._ftol = ftol
        self._xtol = xtol
        self._max_nfev = max_nfev
        self._x_scale = x_scale
        self._tr_solver = tr_solver
        self._include_bounds = include_bounds
        self._num_passes = num_passes

        self._apply_state_scaling = apply_state_scaling

        self._kwargs = kwargs

    def retrieve(
        self,
        measurement_l1: RadianceBase,
        forward_model: ForwardModel,
        retrieval_target: RetrievalTarget,
    ):
        ### Get the prior values
        x_a = retrieval_target.apriori_state()
        initial_guess = retrieval_target.state_vector()
        lb = retrieval_target.lower_bound()
        ub = retrieval_target.upper_bound()

        inv_Sa = retrieval_target.inverse_apriori_covariance()

        if inv_Sa is None:
            # No apriori covariance/regularization
            # Use initial guess to make the matrices as x_a might be None as well
            inv_Sa = np.zeros((len(initial_guess), len(initial_guess)))

        if x_a is None:
            x_a = np.zeros_like(initial_guess)

        ### Get the measurement values
        y_meas_dict = retrieval_target.measurement_vector(measurement_l1)

        y_meas = y_meas_dict["y"]
        good_meas = ~np.isnan(y_meas)
        y_meas = y_meas[good_meas]

        if "y_error" in y_meas_dict:
            # Have measurement error
            if len(np.shape(y_meas_dict["y_error"])) == 1:
                # Only supplied is the diagonal of the error elements
                Sy = sparse.csc_matrix(
                    sparse.diags(y_meas_dict["y_error"][good_meas], 0)
                )
                inv_Sy = sparse.csc_matrix(
                    sparse.diags(1 / y_meas_dict["y_error"][good_meas], 0)
                )
                y_scaler_inv = np.linalg.cholesky(inv_Sy)
            elif sparse.issparse(y_meas_dict["y_error"]):
                Sy = y_meas_dict["y_error"][np.ix_(good_meas, good_meas)]

                rows = Sy.indices  # Row indices of non-zero elements
                cols = np.repeat(
                    np.arange(Sy.shape[1]), np.diff(Sy.indptr)
                )  # Column indices of non-zero elements

                # Check if all non-zero elements are on the diagonal
                is_diagonal = np.all(rows == cols)

                if is_diagonal:
                    # Ensure there are no zeros on the diagonal
                    if np.any(Sy.data == 0):
                        msg = "Cannot invert a matrix with zeros on the diagonal."
                        raise ZeroDivisionError(msg)

                    # Compute the inverse of the diagonal elements
                    inv_data = 1.0 / Sy.data

                    # Create the inverse matrix using the same indices and indptr
                    inv_Sy = sparse.csc_matrix(
                        (inv_data, Sy.indices, Sy.indptr), shape=Sy.shape
                    )

                    y_scaler_inv = sparse.csc_matrix(
                        (np.sqrt(inv_data), Sy.indices, Sy.indptr), shape=Sy.shape
                    )
            else:
                Sy = y_meas_dict["y_error"][np.ix_(good_meas, good_meas)]
                inv_Sy = np.linalg.inv(Sy)
                y_scaler_inv = np.linalg.cholesky(inv_Sy)
        else:
            # No user supplied error, use identity matrix
            Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))
            inv_Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))

            y_scaler_inv = np.linalg.cholesky(inv_Sy)

        try:
            chol_inv_Sa = np.linalg.cholesky(inv_Sa)
        except np.linalg.LinAlgError:
            # If the inverse covariance is not positive definite, then we can't use the cholesky
            # decomposition, but we can use an eigenvalue decomposition
            eigvals, eigvecs = np.linalg.eigh(inv_Sa)
            eigvals[eigvals < 0] = 0
            chol_inv_Sa = np.diag(np.sqrt(eigvals)) @ eigvecs.T

        if self._apply_state_scaling:
            x_scaler_inv = np.diag(1 / x_a)
            x_scaler = np.diag(x_a)
        else:
            x_scaler_inv = np.eye(len(x_a))
            x_scaler = np.eye(len(x_a))

        x_a = x_scaler_inv @ x_a

        def residual_fun(x):
            retrieval_target.update_state(x_scaler @ x)

            y_ret_dict = retrieval_target.measurement_vector(
                forward_model.calculate_radiance()
            )

            K = y_ret_dict["jacobian"]

            K = y_scaler_inv @ K[good_meas, :] @ x_scaler
            y_ret = y_scaler_inv @ y_ret_dict["y"][good_meas]

            # First part of residuals is from y, y_meas - y_ret, and jacobian K
            res = y_ret - y_scaler_inv @ y_meas
            # Second part of residuals is x-x_a, with identity jacobian in scaled space
            res_x = chol_inv_Sa @ x_scaler @ (x - x_a)
            K_x = chol_inv_Sa @ x_scaler

            # To match the cost of the standard "Rodgers" minimizer we have to scale by the number of measurements,
            # and also multiply by 2 since the scipy least squares does 0.5 * res.T @ res
            n = len(res) / 2

            return np.concatenate((res, res_x)) / np.sqrt(n), np.vstack(
                [K, K_x]
            ) / np.sqrt(n)

        fun = MemoizeJac(residual_fun)
        jac = fun.derivative

        bounds = (
            (x_scaler_inv @ lb, x_scaler_inv @ ub)
            if self._include_bounds
            else (np.ones_like(lb) * (-np.inf), np.ones_like(ub) * np.inf)
        )

        results = {}

        for _ in range(self._num_passes):
            results["minimizer"] = least_squares(
                fun,
                x0=x_scaler_inv @ initial_guess,
                jac=jac,
                x_scale=self._x_scale,
                verbose=2,
                tr_solver=self._tr_solver,
                max_nfev=self._max_nfev,
                tr_options={"regularize": False},
                method=self._method,
                xtol=self._xtol,
                ftol=self._ftol,
                bounds=bounds,
                **self._kwargs,
            )

            y_ret_dict = retrieval_target.measurement_vector(
                forward_model.calculate_radiance()
            )

            meas_resid = y_meas - y_ret_dict["y"][good_meas]

            median_resid = np.median(np.abs(meas_resid))
            # Adjust the scaler based on the residual fractions
            scaler = np.abs(meas_resid) / median_resid
            scaler[scaler < 1] = 1

            # Rescale the measurement errrors
            y_scaler_inv = sparse.diags(np.sqrt(inv_Sy.diagonal() / scaler**2))

            initial_guess = retrieval_target.state_vector()

        K = y_ret_dict["jacobian"][good_meas, :]

        results.update(estimate_error(K, Sy, inv_Sy, inv_Sa))

        return results


class SciPyMinimizerGrad(Minimizer):
    def __init__(self) -> None:
        super().__init__()

    def retrieve(
        self,
        measurement_l1: RadianceBase,
        forward_model: ForwardModel,
        retrieval_target: RetrievalTarget,
    ):
        ### Get the prior values
        x_a = retrieval_target.apriori_state()
        initial_guess = retrieval_target.state_vector()

        inv_Sa = retrieval_target.inverse_apriori_covariance()

        if inv_Sa is None:
            # No apriori covariance/regularization
            # Use initial guess to make the matrices as x_a might be None as well
            inv_Sa = np.zeros((len(initial_guess), len(initial_guess)))

        if x_a is None:
            x_a = np.zeros_like(initial_guess)

        ### Get the measurement values
        y_meas_dict = retrieval_target.measurement_vector(measurement_l1)

        y_meas = y_meas_dict["y"]
        good_meas = ~np.isnan(y_meas)
        y_meas = y_meas[good_meas]

        if "y_error" in y_meas_dict:
            # Have measurement error
            if len(np.shape(y_meas_dict["y_error"])) == 1:
                # Only supplied is the diagonal of the error elements
                Sy = sparse.csc_matrix(
                    sparse.diags(y_meas_dict["y_error"][good_meas], 0)
                )
                inv_Sy = sparse.csc_matrix(
                    sparse.diags(1 / y_meas_dict["y_error"][good_meas], 0)
                )
            else:
                Sy = y_meas_dict["y_error"][np.ix_(good_meas, good_meas)]
                inv_Sy = np.linalg.inv(Sy)
        else:
            # No user supplied error, use identity matrix
            Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))
            inv_Sy = sparse.csc_matrix(sparse.eye(len(y_meas), len(y_meas)))

        y_scaler_inv = sparse.diags(np.sqrt(inv_Sy.diagonal()))

        x_scaler_inv = np.linalg.cholesky(inv_Sa)
        x_scaler = np.linalg.inv(x_scaler_inv)

        y_meas = y_scaler_inv @ y_meas

        x_a = x_scaler_inv @ x_a

        def residual_fun(x):
            retrieval_target.update_state(x_scaler @ x)

            y_ret_dict = retrieval_target.measurement_vector(
                forward_model.calculate_radiance()
            )

            K = y_scaler_inv @ y_ret_dict["jacobian"][good_meas, :] @ x_scaler
            y_ret = y_scaler_inv @ y_ret_dict["y"][good_meas]

            cost = (y_ret - y_meas).T @ (y_ret - y_meas).T + (x - x_a).T @ (x - x_a)

            grad = K.T @ (y_ret - y_meas) + (x - x_a)
            K_x = np.eye(len(x))

            full_K = np.vstack([K, K_x])

            return (cost, 2 * grad), 2 * full_K.T @ full_K

        fun = MemoizeJac(residual_fun)
        hess = fun.derivative

        return minimize(
            fun,
            x0=x_scaler_inv @ initial_guess,
            jac=True,
            hess=hess,
            options={"disp": True, "maxiter": 30},
            method="trust-exact",
        )



from __future__ import annotations

from copy import copy

import numpy as np
import xarray as xr
from scipy.interpolate import UnivariateSpline

from skretrieval.retrieval.statevector import StateVectorElement
from skretrieval.retrieval.tikhonov import (
    two_dim_horizontal_second_deriv,
    two_dim_vertical_first_deriv,
)


class MultiplicativeSpline(StateVectorElement):
    def __init__(
        self,
        num_los: int,
        low_wavelength_nm: float,
        high_wavelength_nm: float,
        num_wv: int,
        s: float,
        order=3,
        min_value=0.1,
        max_value=3,
    ):
        self._wv = np.linspace(
            low_wavelength_nm, high_wavelength_nm, num_wv, endpoint=True
        )
        self._x = np.ones((num_los, len(self._wv)))
        self._low_wavelength_nm = low_wavelength_nm
        self._high_wavelength_nm = high_wavelength_nm
        self._s = s
        self._order = order
        self._min_value = min_value
        self._max_value = max_value

    def state(self) -> np.array:
        return self._x.flatten()

    def lower_bound(self) -> np.array:
        return np.ones_like(self._x.flatten()) * self._min_value

    def upper_bound(self) -> np.array:
        return np.ones_like(self._x.flatten()) * self._max_value

    def update_state(self, x: np.array):
        self._x = x.reshape(self._x.shape)

    def inverse_apriori_covariance(self) -> np.ndarray:
        # TODO: Fix
        gamma = two_dim_horizontal_second_deriv(  # noqa: F841
            self._x.shape[0], self._x.shape[1], factor=0.01
        )
        return np.eye(len(self.state())) * 1e2

    def name(self) -> str:
        return f"spline_{self._low_wavelength_nm}_{self._high_wavelength_nm}"

    def propagate_wf(self, radiance: xr.Dataset) -> xr.Dataset:
        # Calculate the derivative of the spline
        spline_deriv = np.zeros(
            (self._x.shape[0], self._x.shape[1], len(radiance["wavelength"]))
        )

        wv = radiance["wavelength"].to_numpy()
        good = (wv > self._low_wavelength_nm) & (wv < self._high_wavelength_nm)

        for i in range(self._x.shape[0]):
            bx = copy(self._x[i])
            base_spline = UnivariateSpline(self._wv, bx, s=self._s, k=self._order)
            base_vals = base_spline(wv[good])

            for j in range(len(bx)):
                bx[j] += 1e-2
                p_vals = UnivariateSpline(self._wv, bx, s=self._s, k=self._order)(
                    wv[good]
                )
                bx[j] -= 1e-2

                spline_deriv[i, j, good] = (p_vals - base_vals) / 1e-2

        full_deriv = np.zeros(
            (
                spline_deriv.shape[0],
                spline_deriv.shape[1],
                radiance["radiance"].shape[2],
                radiance["radiance"].shape[0],
                radiance["radiance"].shape[1],
            )
        )

        for i in range(self._x.shape[0]):
            full_deriv[i, :, :, :, i] = (
                spline_deriv[i, :, :, np.newaxis]
                * radiance["radiance"].to_numpy()[:, i, :]
            ).transpose([0, 2, 1])

        return xr.DataArray(
            full_deriv.reshape(
                (
                    -1,
                    radiance["radiance"].shape[2],
                    radiance["radiance"].shape[0],
                    radiance["radiance"].shape[1],
                )
            ),
            dims=["x", "stokes", "wavelength", "los"],
        )

    def modify_input_radiance(self, radiance: xr.Dataset):
        wv = radiance["wavelength"].to_numpy()
        vals = np.ones((len(wv), self._x.shape[0]))
        good = (wv > self._low_wavelength_nm) & (wv < self._high_wavelength_nm)

        for i in range(self._x.shape[0]):
            base_spline = UnivariateSpline(
                self._wv, self._x[i], s=self._s, k=self._order
            )
            vals[good, i] = base_spline(wv[good])

        radiance *= xr.DataArray(vals, dims=["wavelength", "los"])

        return radiance

    def apriori_state(self) -> np.array:
        return np.ones_like(self.state())


class MultiplicativeSplineOne(StateVectorElement):
    def __init__(
        self,
        low_wavelength_nm: float,
        high_wavelength_nm: float,
        num_wv: int,
        s: float,
        order=3,
        min_value=-100,
        max_value=100,
    ):
        self._wv = np.linspace(
            low_wavelength_nm, high_wavelength_nm, num_wv, endpoint=True
        )
        self._x = np.ones(len(self._wv))
        self._low_wavelength_nm = low_wavelength_nm
        self._high_wavelength_nm = high_wavelength_nm
        self._s = s
        self._order = order
        self._min_value = min_value
        self._max_value = max_value
        super().__init__(True)

    def state(self) -> np.array:
        return self._x.flatten()

    def lower_bound(self) -> np.array:
        return np.ones_like(self._x.flatten()) * self._min_value

    def upper_bound(self) -> np.array:
        return np.ones_like(self._x.flatten()) * self._max_value

    def update_state(self, x: np.array):
        self._x = x.reshape(self._x.shape)

    def inverse_apriori_covariance(self) -> np.ndarray:
        return np.eye(len(self.state())) * 1e-5

    def name(self) -> str:
        return f"spline_{self._low_wavelength_nm}_{self._high_wavelength_nm}"

    def propagate_wf(self, radiance: xr.Dataset) -> xr.Dataset:
        # Calculate the derivative of the spline
        spline_deriv = np.zeros((len(self._x), len(radiance["wavelength"])))

        wv = radiance["wavelength"].to_numpy()
        good = (wv > self._low_wavelength_nm) & (wv < self._high_wavelength_nm)

        bx = copy(self._x)
        base_spline = UnivariateSpline(self._wv, bx, s=self._s, k=self._order)
        base_vals = base_spline(wv[good])

        for j in range(len(bx)):
            bx[j] += 1e-2
            p_vals = UnivariateSpline(self._wv, bx, s=self._s, k=self._order)(wv[good])
            bx[j] -= 1e-2

            spline_deriv[j, good] = (p_vals - base_vals) / 1e-2

        full_deriv = np.zeros(
            (
                spline_deriv.shape[0],
                radiance["radiance"].shape[0],
                radiance["radiance"].shape[1],
                radiance["radiance"].shape[2],
            )
        )

        for i in range(radiance["radiance"].to_numpy().shape[1]):
            full_deriv[:, :, i, :] = (
                spline_deriv[:, :, np.newaxis]
                * radiance["radiance"].to_numpy()[:, i, :]
            )

        return xr.DataArray(
            full_deriv.reshape(
                (
                    -1,
                    radiance["radiance"].shape[0],
                    radiance["radiance"].shape[1],
                    radiance["radiance"].shape[2],
                )
            ),
            dims=["x", "wavelength", "los", "stokes"],
        )

    def modify_input_radiance(self, radiance: xr.Dataset):
        wv = radiance["wavelength"].to_numpy()
        vals = np.ones(len(wv))
        good = (wv > self._low_wavelength_nm) & (wv < self._high_wavelength_nm)

        base_spline = UnivariateSpline(self._wv, self._x, s=self._s, k=self._order)
        vals[good] = base_spline(wv[good])

        for var in list(radiance):
            if "wavelength" in radiance[var].dims:
                radiance[var] *= xr.DataArray(vals, dims=["wavelength"])

        return radiance

    def apriori_state(self) -> np.array:
        return np.ones_like(self.state())


class AdditiveSpline(StateVectorElement):
    def __init__(
        self,
        num_los: int,
        low_wavelength_nm: float,
        high_wavelength_nm: float,
        num_wv: int,
        s: float,
        order=0,
        min_value=-np.inf,
        max_value=np.inf,
    ):
        self._wv = np.linspace(
            low_wavelength_nm, high_wavelength_nm, num_wv, endpoint=True
        )
        self._x = np.ones((num_los, len(self._wv)))
        self._low_wavelength_nm = low_wavelength_nm
        self._high_wavelength_nm = high_wavelength_nm
        self._s = s
        self._order = order
        self._min_value = min_value
        self._max_value = max_value

    def state(self) -> np.array:
        return self._x.flatten()

    def lower_bound(self) -> np.array:
        return np.ones_like(self._x.flatten()) * self._min_value

    def upper_bound(self) -> np.array:
        return np.ones_like(self._x.flatten()) * self._max_value

    def update_state(self, x: np.array):
        self._x = x.reshape(self._x.shape)

    def inverse_apriori_covariance(self) -> np.ndarray:
        # TODO: Fix
        gamma = two_dim_vertical_first_deriv(  # noqa: F841
            self._x.shape[0], self._x.shape[1], factor=1e6
        )
        return np.eye(len(self.state())) * 1e-20

    def name(self) -> str:
        return f"add_spline_{self._low_wavelength_nm}_{self._high_wavelength_nm}"

    def propagate_wf(self, radiance: xr.Dataset) -> xr.Dataset:
        # Calculate the derivative of the spline
        spline_deriv = np.zeros(
            (self._x.shape[0], self._x.shape[1], len(radiance["wavelength"]))
        )

        wv = radiance["wavelength"].to_numpy()
        good = (wv > self._low_wavelength_nm) & (wv < self._high_wavelength_nm)

        for i in range(self._x.shape[0]):
            bx = copy(self._x[i])
            base_spline = UnivariateSpline(self._wv, bx, s=self._s, k=self._order)
            base_vals = base_spline(wv[good])

            for j in range(len(bx)):
                bx[j] += 1e-2
                p_vals = UnivariateSpline(self._wv, bx, s=self._s, k=self._order)(
                    wv[good]
                )
                bx[j] -= 1e-2

                spline_deriv[i, j, good] = (p_vals - base_vals) / 1e-2

        full_deriv = np.zeros(
            (
                spline_deriv.shape[0],
                spline_deriv.shape[1],
                radiance["radiance"].shape[0],
                radiance["radiance"].shape[1],
                radiance["radiance"].shape[2],
            )
        )

        for i in range(self._x.shape[0]):
            full_deriv[i, :, :, :, i] = spline_deriv[i, :, np.newaxis, :]

        return xr.DataArray(
            full_deriv.reshape(
                (
                    -1,
                    radiance["radiance"].shape[0],
                    radiance["radiance"].shape[1],
                    radiance["radiance"].shape[2],
                )
            ),
            dims=["x", "stokes", "wavelength", "los"],
        )

    def modify_input_radiance(self, radiance: xr.Dataset):
        wv = radiance["wavelength"].to_numpy()
        vals = np.ones((len(wv), self._x.shape[0]))
        good = (wv > self._low_wavelength_nm) & (wv < self._high_wavelength_nm)

        for i in range(self._x.shape[0]):
            base_spline = UnivariateSpline(
                self._wv, self._x[i], s=self._s, k=self._order
            )
            vals[good, i] = base_spline(wv[good])

        radiance["radiance"] += xr.DataArray(vals, dims=["wavelength", "los"])

        return radiance

    def apriori_state(self) -> np.array:
        return np.zeros_like(self.state())



from __future__ import annotations

import numpy as np
import sasktran2 as sk

import skretrieval.retrieval.measvec as mv
from skretrieval.core.lineshape import DeltaFunction
from skretrieval.core.sasktranformat import SASKTRANRadiance
from skretrieval.core.sensor.spectrograph import SpectrographOnlySpectral


def _construct_l1():
    config = sk.Config()

    model_geometry = sk.Geometry1D(
        cos_sza=0.6,
        solar_azimuth=0,
        earth_radius_m=6372000,
        altitude_grid_m=np.arange(0, 65001, 1000),
        interpolation_method=sk.InterpolationMethod.LinearInterpolation,
        geometry_type=sk.GeometryType.Spherical,
    )

    viewing_geo = sk.ViewingGeometry()

    t_alts = np.arange(500, 60000, 1000.0)

    for alt in t_alts:
        ray = sk.TangentAltitudeSolar(
            tangent_altitude_m=alt,
            relative_azimuth=0,
            observer_altitude_m=200000,
            cos_sza=0.6,
        )
        viewing_geo.add_ray(ray)

    wavel = np.arange(280.0, 800.0, 1)
    atmosphere = sk.Atmosphere(
        model_geometry,
        config,
        wavelengths_nm=wavel,
        pressure_derivative=False,
        temperature_derivative=False,
        specific_humidity_derivative=False,
    )

    sk.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk.constituent.Rayleigh()
    atmosphere["ozone"] = sk.climatology.mipas.constituent("O3", sk.optical.O3DBM())
    atmosphere["no2"] = sk.climatology.mipas.constituent(
        "NO2", sk.optical.NO2Vandaele()
    )

    engine = sk.Engine(config, model_geometry, viewing_geo)
    output = engine.calculate_radiance(atmosphere)

    inst_model = SpectrographOnlySpectral(wavel, [DeltaFunction() for _ in wavel])

    l1 = inst_model.model_radiance(SASKTRANRadiance.from_sasktran2(output), None)["I"]

    l1.data = l1.data.drop_vars(["wf_no2_vmr"])
    l1.data = l1.data.rename_vars({"wf_ozone_vmr": "wf"})
    l1.data = l1.data.rename_dims({"ozone_altitude": "x"})

    l1.data.coords["tangent_altitude"] = (["los"], t_alts)
    l1.data = l1.data.set_xindex("tangent_altitude")

    return l1


def test_selector():
    l1 = _construct_l1()

    l1 = mv.pre_process({"measurement": l1})

    mv.select(l1, wavelength=500, tangent_altitude=slice(10000, 40000))



from __future__ import annotations

import logging
import time

import numpy as np


class Timer:
    """
    Simple wrapper to time things easier.
    """

    def __init__(self, name=None):
        self.name = name

    def __enter__(self):
        self.tstart = time.time()

    def __exit__(self, type, value, traceback):
        if self.name:
            logging.info(self.name)
        msg = "Elapsed: " + str(time.time() - self.tstart) + "s"
        logging.info(msg)


def rotation_matrix(axis, angle):
    """
    Create a 3D rotation matrix using the Euler-Rodrigues formula

    :param axis: axis of rotation
    :param angle: angle to rotate around the axis (radians)
    :return: 3x3 rotation matrix
    """

    unit_axis = axis / np.linalg.norm(axis)

    half_angle_rad = angle / 2.0
    sin_half_angle_rad = np.sin(half_angle_rad)
    cos_half_angle_rad = np.cos(half_angle_rad)
    a = cos_half_angle_rad
    b = unit_axis[0] * sin_half_angle_rad
    c = unit_axis[1] * sin_half_angle_rad
    d = unit_axis[2] * sin_half_angle_rad

    R = np.zeros((3, 3))
    R[0, 0] = a * a + b * b - c * c - d * d
    R[0, 1] = 2 * (b * c - a * d)
    R[0, 2] = 2 * (b * d + a * c)
    R[1, 0] = 2 * (b * c + a * d)
    R[1, 1] = a * a + c * c - b * b - d * d
    R[1, 2] = 2 * (c * d - a * b)
    R[2, 0] = 2 * (b * d - a * c)
    R[2, 1] = 2 * (c * d + a * b)
    R[2, 2] = a * a + d * d - b * b - c * c

    return R


def linear_interpolating_matrix(from_grid: np.array, to_grid: np.array):
    M = np.zeros((len(to_grid), len(from_grid)))

    for idx, ele in enumerate(to_grid):
        try:
            idx_above = np.nonzero(from_grid > ele)[0][0]
            interp_ele = ele
            if idx_above == 0:
                idx_above = 1
                interp_ele = from_grid[0]
        except Exception as _:
            idx_above = len(from_grid)
            interp_ele = from_grid[0]

        if idx_above == len(from_grid):
            M[idx, idx_above - 1] = 1
        else:
            w = (from_grid[idx_above] - interp_ele) / (
                from_grid[idx_above] - from_grid[idx_above - 1]
            )
            M[idx, idx_above] = 1 - w
            M[idx, idx_above - 1] = w

    return M


def configure_log():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    class ExFormatter(logging.Formatter):
        def_keys = [  # noqa: RUF012
            "name",
            "msg",
            "args",
            "levelname",
            "levelno",
            "pathname",
            "filename",
            "module",
            "exc_info",
            "exc_text",
            "stack_info",
            "lineno",
            "funcName",
            "created",
            "msecs",
            "relativeCreated",
            "thread",
            "threadName",
            "processName",
            "process",
            "taskName",
            "message",
        ]

        def format(self, record):
            string = super().format(record)
            extra = {k: v for k, v in record.__dict__.items() if k not in self.def_keys}
            if len(extra) > 0:
                string += " - " + str(extra)
            return string

    logger.addHandler(logging.StreamHandler())
    logger.handlers[0].setFormatter(ExFormatter())



from __future__ import annotations

import numpy as np
import sasktran as sk
import xarray as xr

import skretrieval.core.radianceformat as radianceformat
from skretrieval.core import OpticalGeometry
from skretrieval.legacy.core.sensor import Sensor
from skretrieval.retrieval import ForwardModel


class MeasurementSimulator(ForwardModel):
    """
    A simple measurement simulator which acts as a forward model.  This is not a generic class, this is moreso
    an example of what a MeasurementSimulator should do.
    """

    def __init__(
        self,
        sensor: Sensor,
        optical_axis: list[OpticalGeometry],
        atmosphere: sk.Atmosphere,
        options=None,
    ):
        self.sensor = sensor

        self.optical_axis = optical_axis
        self.model_geometry = sk.Geometry()

        self.measurement_geometry = np.hstack(
            [self.sensor.measurement_geometry(o) for o in optical_axis]
        )

        # Determine how to set up the model geometry
        self.model_geometry.lines_of_sight = self.measurement_geometry

        self.engine = sk.EngineHR(
            geometry=self.model_geometry, atmosphere=atmosphere, options=options
        )

        self.model_wavel_nm = self.sensor.required_wavelengths(0.1)

        self.engine.wavelengths = self.model_wavel_nm

    def calculate_radiance(self):
        radiance = self.engine.calculate_radiance()
        if hasattr(radiance, "weighting_function"):
            wf = radiance.weighting_function
            radiance = radiance.radiance
        else:
            wf = None

        model_values = [
            self.sensor.model_radiance(
                o, self.model_wavel_nm, self.model_geometry, radiance, wf
            )
            for o in self.optical_axis
        ]

        # TODO: Figure out how to make the radiance format class do this concatenation rather than the simulator
        if self.sensor.radiance_format() == radianceformat.RadianceGridded:
            data = radianceformat.RadianceGridded(
                xr.concat([m.data for m in model_values], dim="los")
            )
        elif self.sensor.radiance_format() == radianceformat.RadianceRaw:
            data = radianceformat.RadianceRaw(
                xr.concat([m.data for m in model_values], dim="meas")
            )
        else:
            msg = "Simulator does not support the format of the sensor"
            raise ValueError(msg)

        return data



from __future__ import annotations

import abc
from copy import copy
from fnmatch import fnmatch

import numpy as np
import sasktran2 as sk

from skretrieval.core.radianceformat import RadianceGridded


class Observation:
    """
    Abstract base class which defines the interface for an observation
    """

    @abc.abstractmethod
    def sk2_geometry(self, **kwargs) -> dict[sk.ViewingGeometry]:
        """
        The "Ideal" viewing geometry for the observation. One viewing ray for every
        line of sight of the instrument

        Returns
        -------
        dict[sk.ViewingGeometry]
            _description_
        """

    @abc.abstractmethod
    def skretrieval_l1(self, **kwargs) -> dict[RadianceGridded]:
        """
        The L1 data for the observation in the "Core Radiance Format"

        Returns
        -------
        dict[RadianceGridded]
        """

    @abc.abstractmethod
    def sample_wavelengths(self) -> dict[np.array]:
        """
        The sample wavelengths for the observation in [nm]

        Returns
        -------
        dict[np.array]
        """

    @abc.abstractmethod
    def reference_cos_sza(self) -> dict[float]:
        """
        The reference cosine of the solar zenith angle for the observation

        Returns
        -------
        dict[float]
        """

    @abc.abstractmethod
    def reference_latitude(self) -> dict[float]:
        """
        The reference latitude for the observation

        Returns
        -------
        dict[float]
        """

    @abc.abstractmethod
    def reference_longitude(self) -> dict[float]:
        """
        The reference longitude for the observation

        Returns
        -------
        dict[float]
        """

    def append_information_to_l1(self, l1: dict[RadianceGridded], **kwargs) -> None:
        """
        A method that allows for the observation to append information to the L1 data
        simulated by the forward model. Useful for adding things that are in the real L1 data
        to the simulations that may be useful inside the measurement vector.

        Parameters
        ----------
        l1 : dict[RadianceGridded]
        """

    def __add__(self, other: Observation) -> Observation:
        return CombinedObservation(self, other)


class FilteredObservation(Observation):
    def __init__(self, obs: Observation, filter: str):
        """
        An internal class that allows for filtering of observation data.
        Only data that matches the filter will be passed through
        the various methods

        Parameters
        ----------
        obs : Observation
        filter : str
            A string that can be used with fnmatch to filter the keys
        """
        self._obs = obs
        self._filter = filter

    def sk2_geometry(self, **kwargs) -> dict[sk.ViewingGeometry]:
        return {
            k: v
            for k, v in self._obs.sk2_geometry(**kwargs).items()
            if fnmatch(k, self._filter)
        }

    def skretrieval_l1(self, **kwargs) -> dict[RadianceGridded]:
        return {
            k: v
            for k, v in self._obs.skretrieval_l1(**kwargs).items()
            if fnmatch(k, self._filter)
        }

    def sample_wavelengths(self) -> dict[np.array]:
        return {
            k: v
            for k, v in self._obs.sample_wavelengths().items()
            if fnmatch(k, self._filter)
        }

    def reference_cos_sza(self) -> dict[float]:
        return {
            k: v
            for k, v in self._obs.reference_cos_sza().items()
            if fnmatch(k, self._filter)
        }

    def reference_latitude(self) -> dict[float]:
        return {
            k: v
            for k, v in self._obs.reference_latitude().items()
            if fnmatch(k, self._filter)
        }

    def reference_longitude(self) -> dict[float]:
        return {
            k: v
            for k, v in self._obs.reference_longitude().items()
            if fnmatch(k, self._filter)
        }

    def append_information_to_l1(self, l1: dict[RadianceGridded], **kwargs) -> None:
        self._obs.append_information_to_l1(l1, **kwargs)


class CombinedObservation(Observation):
    def __init__(self, obs1: Observation, obs2: Observation):
        """
        An internal class that allows for combining two observations into a single
        observation.

        Parameters
        ----------
        obs1 : Observation
        obs2 : Observation
        """
        self._obs1 = obs1
        self._obs2 = obs2

    def sk2_geometry(self, **kwargs) -> dict[sk.ViewingGeometry]:
        return {
            **self._obs1.sk2_geometry(**kwargs),
            **self._obs2.sk2_geometry(**kwargs),
        }

    def skretrieval_l1(self, *args, **kwargs) -> dict[RadianceGridded]:
        return {
            **self._obs1.skretrieval_l1(*args, **kwargs),
            **self._obs2.skretrieval_l1(*args, **kwargs),
        }

    def sample_wavelengths(self) -> dict[np.array]:
        return {**self._obs1.sample_wavelengths(), **self._obs2.sample_wavelengths()}

    def reference_cos_sza(self) -> dict[float]:
        return {**self._obs1.reference_cos_sza(), **self._obs2.reference_cos_sza()}

    def reference_latitude(self) -> dict[float]:
        return {**self._obs1.reference_latitude(), **self._obs2.reference_latitude()}

    def reference_longitude(self) -> dict[float]:
        return {**self._obs1.reference_longitude(), **self._obs2.reference_longitude()}

    def append_information_to_l1(self, l1: dict[RadianceGridded], **kwargs) -> None:
        self._obs1.append_information_to_l1(l1, **kwargs)
        self._obs2.append_information_to_l1(l1, **kwargs)


class SimulatedObservation(Observation):
    def __init__(
        self,
        geo: sk.ViewingGeometry,
        sample_wavelengths: np.array,
        name="measurement",
        state_adjustment_factors=None,
    ):
        """
        Common funnctionality for ideal simulated observations that consist of a single
        RadianceGridded object

        Parameters
        ----------
        geo : sk.ViewingGeometry
        sample_wavelengths : np.array
        name : str, optional
            Name to use in the measurement dectionaries, by default "measurement"
        state_adjustment_factors : _type_, optional
            Factors {key: val} where state elements indexed by key will be multiplied by val inside the simulation, by default None
        """
        self._geo = geo
        self._sample_wavelengths = sample_wavelengths
        self._name = name
        if state_adjustment_factors is None:
            self._state_adjustment_factors = {}
        else:
            self._state_adjustment_factors = state_adjustment_factors

    def sk2_geometry(self, **kwargs) -> dict[sk.ViewingGeometry]:
        return {self._name: self._geo}

    def _append_information_to_l1(self, l1: dict[RadianceGridded], **kwargs) -> None:
        pass

    def skretrieval_l1(
        self, forward_model, state_vector, l1_kwargs, **kwargs  # noqa: ARG002
    ) -> dict[RadianceGridded]:
        old_x = {}
        for k, v in self._state_adjustment_factors.items():
            old_x[k] = copy(state_vector.sv[k].state())
            if isinstance(v, dict):
                state_vector.sv[k].adjust_constituent_attributes(**v)
            else:
                state_vector.sv[k].update_state(state_vector.sv[k].state() * v)

        l1 = forward_model.calculate_radiance()
        self._append_noise_to_l1(l1)

        for k, _ in self._state_adjustment_factors.items():
            state_vector.sv[k].update_state(old_x[k])

        return l1

    def sample_wavelengths(self) -> np.array:
        return {self._name: self._sample_wavelengths}

    def _append_noise_to_l1(self, l1: dict[RadianceGridded]) -> None:
        for _, v in l1.items():
            v.data["radiance_noise"] = v.data["radiance"] * 0.01


class SimulatedNadirObservation(SimulatedObservation):
    def __init__(
        self,
        cos_sza: float,
        cos_viewing_zenith: float,
        reference_latitude: float,
        reference_longitude: float,
        sample_wavelengths: np.array,
        name: str = "measurement",
        state_adjustment_factors=None,
        noise_fn=None,
    ):
        """
        A simulated nadir observation

        Parameters
        ----------
        cos_sza : float
            Cosine of the solar zenith angle at the ground
        cos_viewing_zenith : float
            Cosone of the viewing zenith angle, 1.0 for pure nadir viewing
        reference_latitude : float
            Latitude of the ground point
        reference_longitude : float
            Longitude of the ground point
        sample_wavelengths : np.array
        name : str, optional
            , by default "measurement"
        state_adjustment_factors : _type_, optional
            , by default None
        """
        self._cos_sza = cos_sza
        self._cos_viewing_zenith = cos_viewing_zenith
        self._reference_latitude = reference_latitude
        self._reference_longitude = reference_longitude
        self._noise_fn = noise_fn

        geo = sk.ViewingGeometry()
        geo.add_ray(sk.GroundViewingSolar(cos_sza, 0, cos_viewing_zenith, 200000))

        super().__init__(
            geo,
            sample_wavelengths,
            name,
            state_adjustment_factors=state_adjustment_factors,
        )

    def reference_cos_sza(self) -> float:
        return {self._name: self._cos_sza}

    def reference_latitude(self) -> float:
        return {self._name: self._reference_latitude}

    def reference_longitude(self) -> float:
        return {self._name: self._reference_longitude}

    def _append_noise_to_l1(self, l1: dict[RadianceGridded]) -> None:
        if self._noise_fn is not None:
            for _, v in l1.items():
                v.data["radiance_noise"] = self._noise_fn(v.data)
        else:
            super()._append_noise_to_l1(l1)


class SimulatedLimbObservation(SimulatedObservation):
    def __init__(
        self,
        cos_sza: float,
        relative_azimuth: float,
        observer_altitude: float,
        reference_latitude: float,
        reference_longitude: float,
        tangent_altitudes: np.array,
        sample_wavelengths: np.array,
        name: str = "measurement",
        state_adjustment_factors=None,
        noise_fn=None,
    ):
        """
        A simulated limb observation

        Parameters
        ----------
        cos_sza : float
            cos of the solar zenith angle for all lines of sight
        relative_azimuth : float
            Relative azimuth angle for all lines of sight
        observer_altitude : float
            Altitude of the observer
        reference_latitude : float
        reference_longitude : float
        tangent_altitudes : np.array
            Tangent altitudes in [m]
        sample_wavelengths : np.array
            Sample wavelengths in [nm]
        name : str, optional
        state_adjustment_factors : _type_, optional
        """
        self._cos_sza = cos_sza
        self._relative_azimuth = relative_azimuth
        self._observer_altitude = observer_altitude
        self._reference_latitude = reference_latitude
        self._reference_longitude = reference_longitude
        self._tan_alts = tangent_altitudes
        self._noise_fn = noise_fn

        geo = sk.ViewingGeometry()
        for tan_alt in tangent_altitudes:
            geo.add_ray(
                sk.TangentAltitudeSolar(
                    tan_alt, relative_azimuth, observer_altitude, cos_sza
                )
            )

        super().__init__(
            geo,
            sample_wavelengths,
            name,
            state_adjustment_factors=state_adjustment_factors,
        )

    def reference_cos_sza(self) -> float:
        return {self._name: self._cos_sza}

    def reference_latitude(self) -> float:
        return {self._name: self._reference_latitude}

    def reference_longitude(self) -> float:
        return {self._name: self._reference_longitude}

    def append_information_to_l1(self, l1: dict[RadianceGridded], **kwargs) -> None:
        for name, v in l1.items():
            if name.startswith(self._name):
                v.data.coords["tangent_altitude"] = (["los"], self._tan_alts)
                v.data = v.data.set_xindex("tangent_altitude")

    def _append_noise_to_l1(self, l1: dict[RadianceGridded]) -> None:
        if self._noise_fn is not None:
            for _, v in l1.items():
                v.data["radiance_noise"] = self._noise_fn(v.data)
        else:
            super()._append_noise_to_l1(l1)



from __future__ import annotations

import abc
from collections import defaultdict
from copy import copy

import numpy as np
import sasktran as sk
from scipy import sparse
from scipy.sparse.linalg import spsolve

from skretrieval.core.radianceformat import RadianceBase
from skretrieval.legacy.tomography.grids import OrbitalPlaneGrid
from skretrieval.retrieval import RetrievalTarget
from skretrieval.retrieval.tikhonov import (
    two_dim_horizontal_second_deriv,
    two_dim_vertical_second_deriv,
)


class TwoDimTarget(RetrievalTarget):
    def __init__(
        self,
        species: list[sk.Species],
        grid: OrbitalPlaneGrid,
        state_vector_space: list[str],
        horiz_tikh_factor=0,
        vert_tikh_factor=0,
        use_apriori_for_bounds=True,
        apriori_influence=None,
    ):
        """
        A generic two dimensional retrieval target.  This class implements the two-dimensional nature of the retrieval,
        derived classes implement the species specific retrieval information.

        Derived classes must implement functions to calculate the measurement vector/Jacobian for singular image/scans,
        as well as functions to determined the upper and lower bounds of the retrieval.

        Parameters
        ----------
        species : List[sk.Species]
            A list of species objects that are to be retrieved
        grid : OrbitalPlaneGrid
            The orbital plane grid that is used to define the retrieval grid
        horiz_tikh_factor : float, optional
            Horizontal tikhonov factor used for a second order tikhonov regularization.  Default 0
        vert_tikh_factor : float, optional
            Vertical tikhonov factor used for a second order tikhonov regularization.  Default 0
        state_vector_space: List[str]
            List of either 'log' or 'linear', 'linear' currently may not work.
        use_apriori_for_bounds: bool, optional
            If true, the retrieval is performed on a full grid and edges are handled by pinning the shape using
            the apriori covariance.
        apriori_influence: List[float]
            Constant inverse covariances for each species
        """
        self._species = species
        self._horiz_tikh_factor = horiz_tikh_factor
        self._vert_tikh_factor = vert_tikh_factor
        self._grid = grid
        self._state_vector_space = state_vector_space
        self._apriori_influence = apriori_influence

        self._use_apriori_for_bounds = use_apriori_for_bounds

        self._initial_state_vector = copy(self.state_vector())

        self._max_change_factor = 10
        self._apriori_changed = True

    def _internal_climatology(self):
        """
        List of climatologies for each species being retrieved
        """
        return [species.climatology for species in self._species]

    def _internal_species_id(self):
        """
        List of species identifiers for each species being retrieved.
        """
        return [species.species for species in self._species]

    def state_vector(self):
        """
        Full state vector
        """
        lowerbound = self._lower_wf_bound()
        upperbound = self._upper_wf_bound()
        grid_alts = self._grid.altitudes
        alt_idx = (grid_alts >= lowerbound) & (grid_alts <= upperbound)

        state_values = []
        for clim, species in zip(
            self._internal_climatology(), self._internal_species_id()
        ):
            state_values.append(copy(clim[species])[:, alt_idx])  # numangle, numalt

        for idspecies in range(len(self._internal_climatology())):
            if self._state_vector_space[idspecies] == "log":
                state_values[idspecies] = np.log(state_values[idspecies])

        return np.concatenate([state_value.flatten() for state_value in state_values])

    def _species_state_vector(self, idspecies):
        """
        State vector for an individual species
        """
        x = self.state_vector()
        len_species = int(len(x) / len(self._internal_climatology()))

        return x[idspecies * len_species : (idspecies + 1) * len_species]

    def state_vector_bounded_mask(self, mask=True):
        """
        A mask the same size as state_vector which is true if the state vector is bounded by the apriori
        """
        numangle = self._grid.numhoriz
        numalt = len(self.state_altitudes())

        x_mask = np.ones((numangle, numalt)) * np.nan

        for grid_index in range(numangle):
            upper_bound_altitude = self._upper_retrieval_bound(grid_index)
            lower_bound_altitude = self._lower_retrieval_bound(grid_index)

            # Bound above
            bounded_altitudes = self.state_altitudes() > upper_bound_altitude
            bounded_to = np.nonzero(bounded_altitudes)[0][0] - 1

            x_mask[grid_index, bounded_altitudes] = grid_index * numalt + bounded_to

            # Bound below
            bounded_altitudes = self.state_altitudes() < lower_bound_altitude

            bounded_to = np.nonzero(bounded_altitudes)[0][-1] + 1

            x_mask[grid_index, bounded_altitudes] = grid_index * numalt + bounded_to

            if self._grid.is_extended(grid_index):
                if grid_index <= self._grid.numextended:
                    x_mask[grid_index, :] = numalt * (
                        self._grid.numextended
                    ) + np.arange(0, len(self.state_altitudes()))
                else:
                    x_mask[grid_index, :] = numalt * (
                        numangle - self._grid.numextended
                    ) + np.arange(0, len(self.state_altitudes()))

        if mask:
            return ~np.isnan(x_mask.flatten())
        return x_mask.flatten()

    def zero_apriori_from_mask(self, gamma, mask=True):
        """
        Takes a tikhonov regularization matrix and zeros out elements that are affected by a bounded state vector
        profile
        """
        bounded_mask = self.state_vector_bounded_mask(mask)

        # Any row of gamma that has contribution from a bounded value should be zero'd out
        row_contrib = np.abs(gamma) @ (bounded_mask).astype(int)
        new_gamma = gamma.asformat("lil")
        new_gamma[row_contrib > 0, :] = 0

        return new_gamma.asformat("csr")

    def scale_apriori(self, gamma, idspecies):
        """
        Scales a tikhonov regularization matrix by the state vector
        """
        if self._state_vector_space[idspecies] == "log":
            return gamma
        x = self._species_state_vector(idspecies)
        x = sparse.diags(1 / x)
        return gamma * x

    def measurement_vector(self, l1_data: RadianceBase):
        output = defaultdict(list)

        for image_idx in range(l1_data.num_images):
            image_l1 = l1_data.image_radiance(image_idx, dense_wf=True)

            image_meas_vec = self._image_measurement_vector(image_l1, image_idx)

            # Don't need the WF anymore
            l1_data.del_wf(image_idx)

            for key, item in image_meas_vec.items():
                if len(item.shape) == 1:
                    output[key].append(item)
                else:
                    output[key].append(sparse.csc_matrix(item))

        y_lengths = [len(y) for y in output["y"]]
        output["y_image_lengths"] = y_lengths

        for key, item in output.items():
            if sparse.issparse(item[0]):
                output[key] = sparse.vstack(item)
            else:
                if hasattr(item[0], "shape") and len(item[0].shape) == 1:
                    output[key] = np.concatenate(item)
                else:
                    output[key] = np.vstack(item)

        if "jacobian" in output:
            diags = []
            for idspecies in range(len(self._internal_climatology())):
                x = np.exp(self._species_state_vector(idspecies))
                if self._state_vector_space[idspecies] == "log":
                    diags.append(x)
                else:
                    diags.append(np.ones_like(x))
            diags = np.concatenate(diags)
            x = sparse.diags(diags)
            output["jacobian"] = (output["jacobian"] * x).asformat("csc")

        return output

    def update_state(self, x: np.ndarray):
        current_state = self.state_vector()
        num_species = len(self._internal_climatology())

        for idspecies, (clim, species, cur_state, new_x) in enumerate(
            zip(
                self._internal_climatology(),
                self._internal_species_id(),
                np.split(current_state, num_species),
                np.split(x, num_species),
            )
        ):
            current_state_resized = cur_state.reshape((self._grid.numhoriz, -1))
            new_state = new_x.reshape((self._grid.numhoriz, -1))

            if self._state_vector_space[idspecies] == "log":
                current_state_resized = np.exp(current_state_resized)
                new_state = np.exp(new_state)

            mult_factor = new_state / current_state_resized

            mult_factor[~np.isfinite(mult_factor)] = 1

            if self._max_change_factor is not None:
                mult_factor[
                    mult_factor > self._max_change_factor
                ] = self._max_change_factor
                mult_factor[mult_factor < 1 / self._max_change_factor] = (
                    1 / self._max_change_factor
                )

            ret_alts = self.state_altitudes()
            all_mult_factors = np.ones_like(clim[species])
            for idx in range(self._grid.numhoriz):
                if not self._use_apriori_for_bounds:
                    good_low_alt = self._lower_retrieval_bound(idx)
                    low_idx = np.nonzero(ret_alts > good_low_alt)[0][0]

                    good_high_alt = self._upper_retrieval_bound(idx)
                    high_idx = np.nonzero(ret_alts > good_high_alt)[0][0]

                    all_mult_factors[idx, :] = np.interp(
                        self._grid.altitudes,
                        ret_alts[low_idx:high_idx],
                        mult_factor[idx, low_idx:high_idx],
                        left=mult_factor[idx, low_idx],
                        right=mult_factor[idx, high_idx],
                    )
                else:
                    all_mult_factors[idx, :] = np.interp(
                        self._grid.altitudes,
                        ret_alts,
                        mult_factor[idx, :],
                        left=mult_factor[idx, 0],
                        right=mult_factor[idx, -1],
                    )
            clim[species] *= all_mult_factors

    def _compute_apriori_parameters(self):
        if self._apriori_changed:
            num_species = len(self._internal_climatology())
            n = int(len(self.state_vector()) / num_species)

            all_S_a = []
            all_x_a = []
            for idspecies in range(num_species):
                inv_S_a = sparse.csc_matrix((n, n))

                numangle = self._grid.numhoriz
                numalt = n // numangle

                if self._horiz_tikh_factor is not None:
                    gamma = two_dim_horizontal_second_deriv(
                        numangle, numalt, self._horiz_tikh_factor, sparse=True
                    )
                    gamma = self.zero_apriori_from_mask(gamma)
                    gamma = self.scale_apriori(gamma, idspecies)

                    inv_S_a += gamma.T @ gamma
                if self._vert_tikh_factor is not None:
                    gamma = two_dim_vertical_second_deriv(
                        numangle, numalt, self._vert_tikh_factor, sparse=True
                    )
                    gamma = self.zero_apriori_from_mask(gamma)
                    gamma = self.scale_apriori(gamma, idspecies)

                    inv_S_a += gamma.T @ gamma

                if self._apriori_influence[idspecies] is not None:
                    gamma = sparse.diags(
                        np.ones(n) * self._apriori_influence[idspecies]
                    )
                    gamma = self.zero_apriori_from_mask(gamma)
                    gamma = self.scale_apriori(gamma, idspecies)

                    inv_S_a += gamma

                if self._use_apriori_for_bounds:
                    inv_S_a_bounding = sparse.csc_matrix((n, n))
                    high_factor = 1e3
                    # Create a very large relative apriori scaling
                    # First order tikhonov matrix is (n+1, n), create this by making two diagonals that are (numangle, numalt)
                    # and then flatten them
                    diagonal = np.ones((numangle, numalt))
                    off_diagonal = np.ones((numangle, numalt)) * -1
                    for grid_index in range(numangle):
                        upper_bound_altitude = self._upper_retrieval_bound(grid_index)
                        lower_bound_altitude = self._lower_retrieval_bound(grid_index)

                        # Bound above
                        bounded_altitudes = (
                            self.state_altitudes() > upper_bound_altitude
                        )
                        first_idx = np.nonzero(bounded_altitudes)[0][0]
                        bounded_altitudes[first_idx - 1] = True
                        diagonal[grid_index, ~bounded_altitudes] = 0
                        off_diagonal[grid_index, ~bounded_altitudes] = 0

                        # Bound below
                        bounded_altitudes = (
                            self.state_altitudes() < lower_bound_altitude
                        )
                        # last_idx = np.nonzero(~bounded_altitudes)[0][0]
                        # bounded_altitudes[last_idx] = True

                        diagonal[grid_index, bounded_altitudes] = 1
                        off_diagonal[grid_index, bounded_altitudes] = -1

                        # Unlink the last point to the next angular point
                        diagonal[grid_index, -1] = 0
                        off_diagonal[grid_index, -1] = 0
                    gamma = (
                        sparse.diags(
                            [diagonal.flatten(), off_diagonal.flatten()], offsets=[0, 1]
                        )
                        * high_factor
                    )
                    gamma = self.scale_apriori(gamma, idspecies)
                    inv_S_a_bounding += gamma.T @ gamma
                    self._vert_gamma = copy(gamma)

                    # Same thing but for horizontal regularization to enforce the edges
                    num_extended = self._grid.numextended
                    if num_extended > 0:
                        diagonal_horiz = np.zeros((numangle, numalt))
                        off_diagonal_horiz = np.zeros((numangle + 1, numalt))

                        # Forward difference from the left edge
                        diagonal_horiz[: num_extended + 1, :] = high_factor
                        off_diagonal_horiz[: num_extended + 1, :] = -high_factor

                        # Forward difference from the right edge
                        diagonal_horiz[-(num_extended + 1) : -1, :] = high_factor
                        off_diagonal_horiz[-(num_extended + 2) :, :] = -high_factor

                        gamma_horiz = sparse.diags(
                            [diagonal_horiz.flatten(), off_diagonal_horiz.flatten()],
                            [0, numalt],
                            shape=((numangle + 1) * numalt, numalt * numangle),
                        )

                        gamma_horiz = self.scale_apriori(gamma_horiz, idspecies)

                        self._horiz_gamma = copy(gamma_horiz)
                        inv_S_a_bounding += gamma_horiz.T @ gamma_horiz

                    full_inv_S_a = inv_S_a + inv_S_a_bounding
                    x_a_bounding = self.state_vector()[
                        (idspecies * n) : (idspecies + 1) * n
                    ]
                    x_a = self._initial_state_vector[
                        (idspecies * n) : (idspecies + 1) * n
                    ]

                    full_x_a = spsolve(
                        full_inv_S_a, inv_S_a_bounding @ x_a_bounding + inv_S_a @ x_a
                    )
                    all_x_a.append(full_x_a)

                    all_S_a.append(full_inv_S_a)

            self._x_a = np.concatenate(all_x_a)
            self._inv_S_a = sparse.block_diag(all_S_a, format="csc")
        self._apriori_changed = False

    def apriori_state(self) -> np.array:
        self._compute_apriori_parameters()
        return self._x_a

    def inverse_apriori_covariance(self):
        self._compute_apriori_parameters()
        return self._inv_S_a

    @abc.abstractmethod
    def _image_measurement_vector(self, l1_data: RadianceBase, image_index):
        """
        Calculate the measurement vector for a single image/scan

        Parameters
        ----------
        l1_data : RadianceBase
            l1_data for the image at image_index
        image_index: int
            Index of the image

        Returns
        -------
        dict
            Dictionary with keys 'y', and optionally 'y_error' and 'jacobian'.  Additional diagnostic keys may
            be included but are not used.
        """

    @abc.abstractmethod
    def _lower_wf_bound(self):
        """
        Lower weighting function altitude in m
        """

    @abc.abstractmethod
    def _upper_wf_bound(self):
        """
        Upper weighting function altitude in m
        """

    @abc.abstractmethod
    def _upper_retrieval_bound(self, grid_index):
        """
        Upper altitude in m at horizontal grid index grid_index

        """

    @abc.abstractmethod
    def _lower_retrieval_bound(self, grid_index):
        """
        Lower altitude in m at horizontal grid index grid_index
        """

    def wfwidths_and_alts(self):
        """
        weighting function widths and altitudes that are intended to be included as options for the HR model.
        """
        wf_alts = self.state_altitudes()
        wf_widths = np.gradient(wf_alts)

        return wf_widths, wf_alts

    def state_altitudes(self):
        """
        Altitudes that the weighting functions are calculated at, and thus deltas for the state vector are calculated
        at

        """
        lowerbound = self._lower_wf_bound()
        upperbound = self._upper_wf_bound()

        grid_alts = self._grid.altitudes

        alt_idx = (grid_alts >= lowerbound) & (grid_alts <= upperbound)

        return grid_alts[alt_idx]



from __future__ import annotations

import numpy as np
import sasktran as sk

from skretrieval.core.radianceformat import RadianceBase, RadianceGridded
from skretrieval.retrieval import ForwardModel, RetrievalTarget


class OzoneRetrieval(RetrievalTarget):
    def __init__(self, ozone_species: sk.Species):
        """
        A test species used just for testing, not intended for real use
        """
        self._ozone_species = ozone_species

        self._retrieval_altitudes = np.arange(10500, 60500, 1000)
        self._atmosphere_altitudes = ozone_species.climatology.altitudes

    def initialize(self, forward_model: ForwardModel, meas_l1: RadianceBase):
        return super().initialize(forward_model, meas_l1)

    def measurement_vector(self, l1_data: RadianceBase):
        if not isinstance(l1_data, RadianceGridded):
            msg = "Class OzoneRetrieval only supports data in the form RadianceGridded"
            raise ValueError(msg)

        triplet_values = l1_data.data.sel(wavelength=600, method="nearest")

        result = {}

        result["y"] = triplet_values["radiance"].to_numpy()
        if "wf" in triplet_values:
            result["jacobian"] = triplet_values["wf"].to_numpy()
        return result

    def state_vector(self):
        return self._ozone_species.climatology.get_parameter(
            self._ozone_species.species, 0, 0, self._retrieval_altitudes, 54372
        )

    def update_state(self, x: np.ndarray):
        current_x = self.state_vector()

        mult_factors = x / current_x

        mult_factors = np.interp(
            self._atmosphere_altitudes,
            self._retrieval_altitudes,
            mult_factors,
            left=mult_factors[0],
            right=mult_factors[-1],
        )

        mult_factors[mult_factors < 0.2] = 0.2
        mult_factors[mult_factors > 5] = 5

        self._ozone_species.climatology[self._ozone_species.species] *= mult_factors

    def apriori_state(self):
        return None

    def inverse_apriori_covariance(self):
        return None



from __future__ import annotations

import numpy as np
import pytest
import sasktran2 as sk2

sk = pytest.importorskip("sasktran")

from skretrieval.core import OpticalGeometry  # noqa: E402
from skretrieval.core.lineshape import Gaussian, Rectangle  # noqa: E402
from skretrieval.core.sasktranformat import SASKTRANRadiance  # noqa: E402
from skretrieval.core.sensor.spectrograph import Spectrograph  # noqa: E402
from skretrieval.legacy.util import convert_sasktran_legacy_geometry  # noqa: E402


def test_spectrograph():
    alt_grid = np.arange(0, 65001, 1000.0)

    tanalts_km = np.arange(10, 60, 2.0)

    geo = sk.VerticalImage()

    geo.from_sza_saa(
        sza=60,
        saa=60,
        lat=0,
        lon=0,
        tanalts_km=tanalts_km,
        mjd=54372,
        locallook=0,
        satalt_km=600,
        refalt_km=20,
    )

    new_geo = convert_sasktran_legacy_geometry(geo)

    model_geometry = sk2.Geometry1D(
        np.cos(np.deg2rad(60)),
        0,
        6372000,
        alt_grid,
        sk2.InterpolationMethod.LinearInterpolation,
        sk2.GeometryType.Spherical,
    )

    config = sk2.Config()

    engine = sk2.Engine(config, model_geometry, new_geo)

    atmosphere = sk2.Atmosphere(
        model_geometry, config, wavelengths_nm=np.arange(280, 800, 0.1)
    )

    sk2.climatology.us76.add_us76_standard_atmosphere(atmosphere)

    atmosphere["rayleigh"] = sk2.constituent.Rayleigh()
    atmosphere["o3"] = sk2.climatology.mipas.constituent("O3", sk2.optical.O3DBM())

    radiance = SASKTRANRadiance.from_sasktran2(
        engine.calculate_radiance(atmosphere), collapse_scalar_stokes=True
    )
    radiance.data["look_vectors"] = (
        ["los", "xyz"],
        np.vstack([los.look_vector for los in geo.lines_of_sight]),
    )

    wavel_grid = np.arange(280, 800, 1.0)

    pixel_shape = [Gaussian(fwhm=1) for _ in wavel_grid]

    spectrograph = Spectrograph(
        wavel_grid, pixel_shape, Gaussian(fwhm=1), Rectangle(1, mode="constant")
    )

    optical_geo = OpticalGeometry(
        observer=geo.lines_of_sight[10].observer,
        look_vector=geo.lines_of_sight[10].look_vector,
        mjd=geo.lines_of_sight[10].mjd,
        local_up=geo.lines_of_sight[10].observer
        / np.linalg.norm(geo.lines_of_sight[10].observer),
    )

    _ = spectrograph.model_radiance(radiance=radiance, orientation=optical_geo)



from __future__ import annotations

from copy import copy
from typing import ClassVar

import numpy as np
import sasktran2 as sk2

import skretrieval.retrieval.prior as prior
from skretrieval.retrieval.rodgers import Rodgers
from skretrieval.retrieval.scipy import SciPyMinimizer, SciPyMinimizerGrad
from skretrieval.retrieval.statevector.constituent import StateVectorElementConstituent
from skretrieval.retrieval.statevector.shifts import WavenumberShift
from skretrieval.retrieval.statevector.spline import MultiplicativeSplineOne

from .ancillary import Ancillary, US76Ancillary
from .forwardmodel import ForwardModelHandler, IdealViewingSpectrograph
from .measvec import MeasurementVector, select
from .observation import Observation
from .statevector.altitude import AltitudeNativeStateVector
from .target.mvtarget import MeasVecTarget


class Retrieval:
    _optical_property_fns: ClassVar[dict[str, callable]] = {}
    _prior_fns: ClassVar[dict[str, callable]] = {}
    _state_fns: ClassVar[dict[str, dict[str, callable]]] = {
        "absorbers": {},
        "aerosols": {},
        "splines": {},
        "shifts": {},
        "surface": {},
        "other": {},
    }

    def _context_fn(_):
        return {}

    @classmethod
    def register_context(cls):
        def decorator(context_fn: callable):
            cls._context_fn = context_fn
            return context_fn

        return decorator

    @classmethod
    def register_optical_property(cls, species_name: str):
        def decorator(optical_property_fn: callable):
            cls._optical_property_fns[species_name] = optical_property_fn
            return optical_property_fn

        return decorator

    @classmethod
    def register_prior(cls, species_name: str):
        def decorator(prior_fn: callable):
            cls._prior_fns[species_name] = prior_fn
            return prior_fn

        return decorator

    @classmethod
    def register_state(cls, category: str, species_name: str):
        def decorator(state_fn: callable):
            cls._state_fns[category][species_name] = state_fn
            return state_fn

        return decorator

    def __init__(
        self,
        observation: Observation,
        measvec: dict[MeasurementVector] | None = None,
        forward_model_cfg: dict | None = None,
        minimizer="rodgers",
        ancillary: Ancillary | None = None,
        l1_kwargs: dict | None = None,
        model_kwargs: dict | None = None,
        minimizer_kwargs: dict | None = None,
        target_kwargs: dict | None = None,
        state_kwargs: dict | None = None,
        **kwargs,
    ) -> None:
        """
        The main processing script that handles the retrieval

        Parameters
        ----------
        observation : Observation
        measvec : dict[MeasurementVector] | None, optional
            Measurement vectors to use, by default will use the raw radiances, by default None
        minimizer : str, optional
            Selects which minimizer to use, default is "rodgers", by default "rodgers"
        l1_kwargs : dict | None, optional
            Additional arguments passed to the observation when constructing the L1, by default None
        model_kwargs : dict | None, optional
            Additional arguments passed to the SASKTRAN2 engine, by default None
        minimizer_kwargs : dict | None, optional
            Additional arguments passed to the minimizer, by default None
        target_kwargs : dict | None, optional
            Additional arguments passed to the retrieval target, by default None
        state_kwargs : dict | None, optional
            Arguments to construct the state vector, by default None
        forward_model_cfg : dict | None, optional
            Additional arguments passed to the forward model, by default None
        """
        if minimizer.lower() == "rodgers":
            # Override the default Rodgers options
            self._minimizer_kwargs = {
                "lm_damping_method": "fletcher",
                "lm_damping": 0.1,
                "max_iter": 30,
                "lm_change_factor": 2,
                "iterative_update_lm": True,
                "retreat_lm": False,
                "apply_cholesky_scaling": True,
                "convergence_factor": 1e-2,
                "convergence_check_method": "dcost",
            }
        else:
            self._minimizer_kwargs = {}
        if minimizer_kwargs is not None:
            self._minimizer_kwargs.update(minimizer_kwargs)

        if state_kwargs is None:
            state_kwargs = {}
        if target_kwargs is None:
            target_kwargs = {}
        if model_kwargs is None:
            model_kwargs = {}
        if l1_kwargs is None:
            l1_kwargs = {}
        if forward_model_cfg is None:
            forward_model_cfg = {"*": {"class": IdealViewingSpectrograph}}
        self._options = kwargs
        self._l1_kwargs = l1_kwargs
        self._minimizer = minimizer
        self._target_kwargs = target_kwargs
        self._state_kwargs = state_kwargs
        self._model_kwargs = model_kwargs
        self._measurement_vector = measvec
        self._forward_model_cfg = forward_model_cfg

        self._ancillary = ancillary

        self._measurement_vector = self._construct_measurement_vector()

        self._observation = observation

        self._anc = self._construct_ancillary()

        self._context = self._context_fn()

        self._state_vector = self._construct_state_vector()
        self._forward_model = self._construct_forward_model()
        self._target = self._construct_target()

        self._obs_l1 = self._observation.skretrieval_l1(
            self._forward_model, self._state_vector, self._l1_kwargs
        )

    def _construct_measurement_vector(self):
        if self._measurement_vector is not None:
            return self._measurement_vector
        return {
            "measurement": MeasurementVector(
                lambda l1, ctxt, **kwargs: select(l1, **kwargs)  # noqa: ARG005
            )
        }

    def _construct_forward_model(self):
        engine_config = sk2.Config()

        for k, v in self._model_kwargs.items():
            setattr(engine_config, k, v)

        return ForwardModelHandler(
            self._forward_model_cfg,
            self._observation,
            self._state_vector,
            self._measurement_vector,
            self._anc,
            engine_config,
        )

    def _const_from_mipas(
        self,
        alt_grid,
        species_name,
        optical,
        prior_infl=1e-2,
        tikh=1e8,
        log_space=False,
        min_val=0,
        max_val=1,
    ):
        const = sk2.climatology.mipas.constituent(species_name, optical)

        altitudes_m = const._constituent.altitudes_m

        new_vmr = np.interp(alt_grid, altitudes_m, const.vmr)

        new_const = sk2.constituent.VMRAltitudeAbsorber(
            optical, alt_grid, new_vmr, out_of_bounds_mode="extend"
        )

        min_val = 1e-40 if log_space else min_val

        return StateVectorElementConstituent(
            new_const,
            species_name,
            ["vmr"],
            min_value={"vmr": min_val},
            max_value={"vmr": max_val},
            prior={
                "vmr": tikh * prior.VerticalTikhonov(1, new_vmr)
                + prior_infl * prior.ConstantDiagonalPrior()
            },
            log_space=log_space,
        )

    def _optical_property(self, species_name: str):
        return self._optical_property_fns[species_name]()

    @staticmethod
    def _default_state_absorber(self, name: str, native_alt_grid: np.array, cfg: dict):
        const = self._const_from_mipas(
            native_alt_grid,
            name,
            self._optical_property(name),
            tikh=cfg["tikh_factor"],
            prior_infl=cfg["prior_influence"],
            log_space=cfg["log_space"],
            min_val=cfg["min_value"],
            max_val=cfg["max_value"],
        )

        const.enabled = cfg.get("enabled", True)
        return const

    @staticmethod
    def _default_state_surface(
        self, name: str, native_alt_grid: np.array, cfg: dict  # noqa: ARG004
    ):
        msg = f"Surface {name} does not have a default implementation"
        raise ValueError(msg)

    @staticmethod
    def _default_state_spline(
        self, name: str, native_alt_grid: np.array, cfg: dict  # noqa: ARG004
    ):
        msg = f"Spline {name} does not have a default implementation"
        raise ValueError(msg)

    @staticmethod
    def _default_state_aerosol(
        self, name: str, native_alt_grid: np.array, cfg: dict  # noqa: ARG004
    ):
        msg = f"aerosol {name} does not have a default implementation"
        raise ValueError(msg)

    @staticmethod
    def _default_state_shift(
        self, name: str, native_alt_grid: np.array, cfg: dict  # noqa: ARG004
    ):
        msg = f"shift {name} does not have a default implementation"
        raise ValueError(msg)

    def _construct_state_vector(self):
        native_alt_grid = self._state_kwargs["altitude_grid"]

        absorbers = {}

        for name, options in self._state_kwargs.get("absorbers", {}).items():
            absorbers[name] = self._state_fns["absorbers"].get(
                name, self._default_state_absorber
            )(self, name, native_alt_grid, options)

        surface = {}
        for name, options in self._state_kwargs.get("surface", {}).items():
            surface[name] = self._state_fns["surface"].get(
                name, self._default_state_surface
            )(self, name, native_alt_grid, options)

        aerosols = {}
        for name, aerosol in self._state_kwargs.get("aerosols", {}).items():
            aerosols[f"{name}"] = self._state_fns["aerosols"].get(
                aerosol["type"], self._default_state_aerosol
            )(self, name, native_alt_grid, aerosol)

        splines = {}
        for name, spline in self._state_kwargs.get("splines", {}).items():
            splines[name] = self._state_fns["splines"].get(
                name, self._default_state_spline
            )(self, name, native_alt_grid, spline)

        shifts = {}
        for name, shift in self._state_kwargs.get("shifts", {}).items():
            shifts[name] = self._state_fns["shifts"].get(
                shift["type"], self._default_state_shift
            )(self, name, native_alt_grid, shift)

        return AltitudeNativeStateVector(
            native_alt_grid, **absorbers, **surface, **aerosols, **splines, **shifts
        )

    def _construct_ancillary(self):
        if self._ancillary is None:
            return US76Ancillary()
        return self._ancillary

    def _construct_target(self):
        return MeasVecTarget(
            self._state_vector,
            self._measurement_vector,
            self._context,
            **self._target_kwargs,
        )

    def _construct_output(self, rodgers_output: dict):
        return rodgers_output

    def retrieve(
        self,
        enabled_state_elements: list[str] | None = None,
        enabled_measurement_vectors: list[str] | None = None,
    ):
        if self._minimizer == "rodgers":
            minimizer = Rodgers(**self._minimizer_kwargs)
        elif self._minimizer == "scipy":
            minimizer = SciPyMinimizer(**self._minimizer_kwargs)
        elif self._minimizer == "scipy_grad":
            minimizer = SciPyMinimizerGrad()

        if enabled_state_elements is not None:
            for key, val in self._state_vector.sv.items():
                if key in enabled_state_elements:
                    val.enabled = True
                else:
                    val.enabled = False

        if enabled_measurement_vectors is not None:
            for key, val in self._measurement_vector.items():
                if key in enabled_measurement_vectors:
                    val.enabled = True
                else:
                    val.enabled = False

        self._target.update_state_slices()

        min_results = minimizer.retrieve(
            self._obs_l1, self._forward_model, self._target
        )

        # Reset the enabled flag
        for _, val in self._state_vector.sv.items():
            val.enabled = True

        for _, val in self._measurement_vector.items():
            val.enabled = True

        # Post process
        final_l1 = self._forward_model.calculate_radiance()
        meas_l1 = self._obs_l1

        results = {}

        results["minimizer"] = min_results
        results["meas_l1"] = meas_l1
        results["simulated_l1"] = final_l1

        results["state"] = self._state_vector.describe(min_results)

        return self._construct_output(results)


# Register all the default optical properties
@Retrieval.register_optical_property("o3")
def o3_optical_property(*args, **kwargs):
    return sk2.optical.O3DBM()


@Retrieval.register_optical_property("no2")
def no2_optical_property(*args, **kwargs):
    return sk2.optical.NO2Vandaele()


@Retrieval.register_optical_property("bro")
def bro_optical_property(*args, **kwargs):
    return sk2.optical.HITRANUV("BrO")


@Retrieval.register_optical_property("so2")
def so2_optical_property(*args, **kwargs):
    return sk2.optical.HITRANUV("SO2")


# Register the default Lambertian surface state
@Retrieval.register_state("surface", "lambertian_albedo")
def lambertian_state(self, name, native_alt_grid: np.array, cfg: dict):  # noqa: ARG001
    albedo_wavel = cfg["wavelengths"]
    albedo_start = np.ones(len(albedo_wavel)) * cfg["initial_value"]

    albedo_const = sk2.constituent.LambertianSurface(
        albedo_start, albedo_wavel, cfg.get("out_of_bounds_mode", "extend")
    )
    sv_ele = StateVectorElementConstituent(
        albedo_const,
        name,
        ["albedo"],
        min_value={"albedo": 0},
        max_value={"albedo": 1},
        prior={
            "albedo": cfg["tikh_factor"] * prior.VerticalTikhonov(1)
            + cfg["prior_influence"] * prior.ConstantDiagonalPrior()
        },
        log_space=False,
    )
    sv_ele.enabled = cfg.get("enabled", True)

    return sv_ele


@Retrieval.register_state("aerosols", "extinction_profile")
def aerosol_extinction_profile(self, name: str, native_alt_grid: np.array, cfg: dict):
    if cfg.get("prior_state") is not None and False:
        ext = cfg["prior_state"]
    else:
        aero_const = sk2.test_util.scenarios.test_aerosol_constituent(native_alt_grid)

        ext = copy(aero_const.extinction_per_m)

    low_boundary = np.nonzero(ext)[0][0]

    ext[:low_boundary] = ext[low_boundary]

    ext[ext < 1e-15] = 1e-15

    scale_factor = cfg.get("scale_factor", 1)

    secondary_kwargs = {
        name: np.ones_like(native_alt_grid) * cfg["prior"][name]["value"]
        for name in cfg["prior"]
        if name != "extinction_per_m"
    }

    db = self._optical_property(name)

    aero_const = sk2.constituent.ExtinctionScatterer(
        db,
        native_alt_grid,
        ext * scale_factor,
        cfg.get("nominal_wavelength", 745),
        "extend",
        **secondary_kwargs,
    )

    sv_ele = StateVectorElementConstituent(
        aero_const,
        f"{name}",
        cfg["retrieved_quantities"].keys(),
        min_value={
            name: val["min_value"] for name, val in cfg["retrieved_quantities"].items()
        },
        max_value={
            name: val["max_value"] for name, val in cfg["retrieved_quantities"].items()
        },
        prior={
            name: val["tikh_factor"]
            * prior.VerticalTikhonov(
                1, prior_state=secondary_kwargs.get(name, ext * scale_factor)
            )
            + val["prior_influence"] * prior.ConstantDiagonalPrior()
            for name, val in cfg["retrieved_quantities"].items()
        },
        log_space=False,
    )

    sv_ele.enabled = cfg.get("enabled", True)

    if cfg.get("initial_guess") is not None:
        sv_ele.update_state(cfg["initial_guess"])

    return sv_ele


@Retrieval.register_state("shifts", "wavenumber_shift")
def wavenumber_shift(
    self, name: str, native_alt_grid: np.array, cfg: dict  # noqa: ARG001
):
    sv_ele = WavenumberShift(
        cfg["num_los"],
        numerical_delta=cfg.get("numerical_delta", 1e-5),
        tikh_factor=cfg.get("tikh_factor", 1e4),
        prior_factor=cfg.get("prior_factor", 0),
        min_shift=cfg.get("min_shift", -0.1),
        max_shift=cfg.get("max_shift", 0.1),
        apply_to_measurement=cfg.get("apply_to_measurement"),
    )

    sv_ele.enabled = cfg.get("enabled", True)

    return sv_ele


@Retrieval.register_state("splines", "constant_los")
def constant_los(self, name: str, native_alt_grid: np.array, cfg: dict):  # noqa: ARG001
    return MultiplicativeSplineOne(
        cfg["low_wavelength_nm"],
        cfg["high_wavelength_nm"],
        cfg["num_wv"],
        cfg.get("s", 1),
        order=cfg["order"],
        min_value=cfg.get("min_value", 0.5),
        max_value=cfg.get("max_value", 1.5),
    )



from __future__ import annotations

import numpy as np
import sasktran as sk
import sasktranif.sasktranif as skif
from sasktran.climatology import ClimatologyBase

from skretrieval.legacy.tomography.grids import OrbitalPlaneGrid


class OrbitalPlaneClimatology(ClimatologyBase):
    def __init__(self, grid: OrbitalPlaneGrid, values: dict):
        """
        2D Climatology constructed out of a grid representing an orbital plane

        Parameters
        ----------
        grid : OrbitalPlaneGrid
            The grid object defining the orbital plane
        values : dict
            Dictionary with keys being the species names, and the values being 2D arrays of (numhoriz, numalt) where
            numhoriz is the number of horizontal steps in grid, and numalt is the number of altitude steps in grid
        """
        self._grid = grid
        self._values = values

    def supported_species(self):
        return list(self._values.keys())

    def skif_object(self, **kwargs) -> skif.ISKClimatology:
        reference_point = kwargs["engine"].model_parameters["referencepoint"]

        geo = sk.Geodetic()
        geo.from_lat_lon_alt(reference_point[0], reference_point[1], reference_point[2])

        local_angles, angleidx, normalandreference = self._grid.get_local_plane(
            geo.location
        )

        local_values = {}
        for key, item in self._values.items():
            local_values[key] = item[angleidx, :]

        user_clim = sk.ClimatologyUserDefined2D(
            np.rad2deg(local_angles),
            self._grid.altitudes,
            local_values,
            normalandreference[3:],
            normalandreference[:3],
        )

        for _, item in local_values.items():
            if len(local_angles) * len(self._grid.altitudes) != len(item.flatten()):
                raise ValueError()

        return user_clim.skif_object()

    def __getitem__(self, item):
        return self._values[item]

    def __setitem__(self, item, value):
        self._values[item] = np.asarray(value)


class OrbitalPlaneAlbedo(sk.BRDF):
    def __init__(self, grid: OrbitalPlaneGrid):
        self._values = np.zeros(grid._numretprof)
        self._grid = grid

    def skif_object(self, **kwargs):
        """
        Returns the internel SasktranIF object
        """
        reference_point = kwargs["engine"].model_parameters["referencepoint"]

        geo = sk.Geodetic()
        geo.from_lat_lon_alt(reference_point[0], reference_point[1], reference_point[2])

        local_angles, angleidx, normalandreference = self._grid.get_local_plane(
            geo.location
        )

        user_clim = sk.ClimatologyUserDefined2D(
            np.rad2deg(local_angles),
            np.array([0.0, 100000.0]),
            {
                "SKCLIMATOLOGY_ALBEDO": np.tile(
                    self._values[angleidx], (2, 1)
                ).transpose()
            },
            normalandreference[3:],
            normalandreference[:3],
        )

        self._brdf = sk.BRDF("PLANE")
        self._brdf.skif_object().SetProperty("clim", user_clim.skif_object())
        return self._brdf.skif_object()

    @property
    def values(self):
        return self._values

    @values.setter
    def values(self, values):
        self._values = values



from __future__ import annotations

import abc
from copy import copy

import numpy as np
import sasktran as sk
import xarray as xr

from skretrieval.retrieval.statevector import StateVectorElement
from skretrieval.retrieval.tikhonov import (
    two_dim_vertical_first_deriv,
    two_dim_vertical_second_deriv,
)
from skretrieval.util import linear_interpolating_matrix


class StateVectorProfile(StateVectorElement):
    def __init__(self, species_name: str, optical_property: sk.OpticalProperty):
        """
        A state vector component that represents a profile of a scattering or absorbing species within the atmosphere.
        This is a base class which other classes can derive


        Parameters
        ----------
        species_name: str
            Name of the species
        optical_property: sk.OpticalProperty
            Optical property of the species
        """
        self._species_name = species_name

        self._optical_property = optical_property
        self._species = None
        self._internal_update()

    def _internal_update(self):
        """
        Updates the internal climatology
        """
        alts, density = self._climatology_altitudes_and_density()
        if self._species is None:
            self._species = sk.Species(
                self._optical_property,
                sk.ClimatologyUserDefined(alts, {self._species_name: density}),
                self._species_name,
            )
        else:
            self._species.climatology[self._species_name] = density

    @abc.abstractmethod
    def _climatology_altitudes_and_density(self):
        pass

    def name(self) -> str:
        return self._species_name

    def add_to_atmosphere(self, atmosphere: sk.Atmosphere):
        """
        Adds the species to the atmosphere

        Parameters
        ----------
        atmosphere: sk.Atmosphere

        """
        atmosphere[self._species_name] = self._species

        if atmosphere.wf_species is None:
            atmosphere.wf_species = [self._species_name]

        if self._species_name not in atmosphere.wf_species:
            old_wf_species = copy(atmosphere.wf_species)
            old_wf_species.append(self._species_name)
            atmosphere.wf_species = old_wf_species


class StateVectorProfilePPM(StateVectorProfile):
    def propagate_wf(self, radiance) -> np.ndarray:
        interp_matrix = linear_interpolating_matrix(
            self._altitudes_m, self._species.climatology.altitudes
        )

        # We have dI/dn, but we need dI/dx, so we have to sum over the deltas that each ppm delta introduces
        backgroundstate = np.asarray(
            self._background_climatology["SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3"]
        )

        wf_name = "wf_" + self.name()
        if wf_name not in radiance:
            msg = "WF not calculated"
            raise ValueError(msg)

        # inlclude ppm conversion here
        new_wf = radiance[wf_name] * backgroundstate / 1e6
        return new_wf @ xr.DataArray(
            data=interp_matrix,
            dims=["perturbation", "x"],
            coords={"perturbation": radiance["perturbation"].to_numpy()},
        )

    def apriori_state(self) -> np.array:
        return self._prior_ppm

    def inverse_apriori_covariance(self) -> np.ndarray:
        return np.linalg.inv(self._prior_ppm_covar)

    def __init__(
        self,
        altitudes_m: np.array,
        values_ppm: np.array,
        species_name: str,
        optical_property: sk.OpticalProperty,
        background_climatology: sk.ClimatologyUserDefined,
        prior_ppm: np.array,
        priori_ppm_covar: np.ndarray,
    ):
        """
        State vector element which defines a vertical profile based on a secondary vertical profile of ppm values.
        The secondary profile can be on any vertical grid (coarser or equal to the regular atmospheric grid, not
        finer)

        Parameters
        ----------
        altitudes_m: np.array
            Altitudes of the PPM grid
        values_ppm: np.array
            ppm of the species
        species_name: np.array
            Name of the species within the atmosphere
        optical_property: sk.OpticalProperty
            Optical property of the species in question
        background_climatology: sk.Climatology
            Climatology which contains a key 'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3' that is the background atmospheric
            number density
        prior_ppm: np.array
            Apriori ppm values for the profile
        priori_ppm_covar: np.ndarray
            Apriori covariance matrix
        """
        self._altitudes_m = np.atleast_1d(altitudes_m)
        self._values_ppm = np.atleast_1d(values_ppm)

        self._prior_ppm = prior_ppm
        self._prior_ppm_covar = priori_ppm_covar

        self._background_climatology = background_climatology

        super().__init__(species_name, optical_property)

    def _climatology_altitudes_and_density(self):
        return self._background_climatology.altitudes, self._convert_to_numden()

    def _convert_to_numden(self):
        background_values = self._background_climatology[
            "SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3"
        ]
        background_altitudes = self._background_climatology.altitudes

        species_fraction = self._values_ppm / 1e6  # Convert from ppm

        interp_fraction = np.interp(
            background_altitudes, self._altitudes_m, species_fraction
        )

        return np.asarray(interp_fraction * background_values)

    def state(self) -> np.array:
        return copy(self._values_ppm)

    def update_state(self, x: np.array):
        self._values_ppm = copy(np.atleast_1d(x))
        self._internal_update()


class StateVectorProfileScale(StateVectorProfile):
    def propagate_wf(self, radiance) -> np.ndarray:
        interp = np.ones((len(self._species.climatology.altitudes), 1))

        wf_name = "wf_" + self.name()

        old_wf = radiance[wf_name]

        return (
            old_wf * self._values[np.newaxis, :] * self._scale_factor
        ) @ xr.DataArray(
            data=interp,
            dims=["perturbation", "x"],
            coords={"perturbation": old_wf["perturbation"].to_numpy()},
        )

    def __init__(
        self,
        scale_factor: float,
        species_name: str,
        optical_property: sk.OpticalProperty,
        altitudes_m: np.array,
        values: np.array,
        scale_factor_sigma: float | None = None,
        scale_factor_prior: float = 1.0,
    ):
        """
        State vector element which defines a vertical profile based on a single scale factor of another profile

        Parameters
        ----------
        scale_factor: float
            Scale factor
        species_name: str
            Name of the species within the atmosphere object
        optical_property: sk.OpticalProperty
            Optical property of the species
        altitudes_m: np.array
            Array of altitudes that define the atmospheric grid
        values: np.array
            Profile values of the underlying profile that is to be scaled
        scale_factor_sigma: float, optional
            1sigma prior value, defualt is None which indicates no prior
        scale_factor_prior: float, optional
            Prior value on the scale factor, default is 1
        """
        self._scale_factor = scale_factor
        self._scale_factor_prior = scale_factor_prior
        self._scale_factor_sigma = scale_factor_sigma
        self._altitudes_m = altitudes_m
        self._values = values

        super().__init__(species_name, optical_property)

    def _climatology_altitudes_and_density(self):
        return self._altitudes_m, self._values * self._scale_factor

    def state(self) -> np.array:
        return [copy(self._scale_factor)]

    def update_state(self, x: np.array):
        self._scale_factor = copy(float(x))
        self._internal_update()

    def described_state(self, stdev: np.array):
        return xr.Dataset(
            {
                f"{self.name()}_scale_factor": float(self._scale_factor),
                f"{self.name()}_scale_factor_stdev": float(stdev),
            }
        )

    def apriori_state(self) -> np.array:
        return [copy(self._scale_factor_prior)]

    def inverse_apriori_covariance(self) -> np.ndarray:
        inv_covar = np.zeros((1, 1), np.float64)
        if self._scale_factor_sigma is not None:
            inv_covar[0, 0] = 1 / (self._scale_factor_sigma**2)

        return inv_covar


class StateVectorProfileLogND(StateVectorProfile):
    def propagate_wf(self, radiance) -> xr.Dataset:
        wf = radiance["wf_" + self.name()]
        new_wf = wf * np.exp(self._values)[np.newaxis, np.newaxis, :]
        return (
            new_wf.isel(perturbation=~self._zero_mask)
            .drop("perturbation")
            .rename({"perturbation": "x"})
        )

    def __init__(
        self,
        altitudes_m: np.array,
        values: np.array,
        species_name: str,
        optical_property: sk.OpticalProperty,
        lowerbound: float,
        upperbound: float,
        second_order_tikhonov_factor: float,
        bounding_factor=1e4,
        zero_factor: float = 1e-20,
        max_update_factor: float = 5,
    ):
        """
        A state vector element which defines a vertical profile as the logarithm of number density.  The profile
        is bounded above and below by the upperbound and lowerbound through the use of the apriori.  A second
        order tikhonov smoothing is included.  Number dennsities less than a certain factor (zero_factor) are treated as
        identically equal to 0 within the RTM to prevent numerical precision problems

        Parameters
        ----------
        altitudes_m: np.array
            Altitudes of the profile, should match weighting function altitudes
        values: np.array
            log(number density) for the profile
        species_name: str
            Name of the species in the atmosphere
        optical_property: sk.OpticalProperty
            Optical property of the species
        lowerbound: float
            Lowerbound of the retrieval in m
        upperbound: float
            Upperbound of the retrieval in m
        second_order_tikhonov_factor: float
            Second order tikhonov factor
        bounding_factor: float, optional
            Bounding factor used for the apriori upper/lower bounding. Default 1e4
        zero_factor: float, optional
            Number densities less than this number are treated as identically 0.  Default 1e-20
        max_update_factor: float, optional
            Maximum factor to update any element of the state vector every iteration. Default 5
        """
        self._altitudes_m = altitudes_m
        self._values = values
        self._zero_factor = zero_factor
        self._zero_mask = np.exp(self._values) < self._zero_factor
        self._lowerbound = lowerbound
        self._upperbound = upperbound
        self._second_order_tikhonov_factor = second_order_tikhonov_factor
        self._bounding_factor = bounding_factor
        self._max_update_factor = max_update_factor

        super().__init__(species_name, optical_property)

        self._initial_state = copy(self.state())
        self._compute_apriori_covariance()

    def _climatology_altitudes_and_density(self):
        values = np.exp(self._values)
        values[self._zero_mask] = 0.0

        return self._altitudes_m, values

    def _compute_apriori_covariance(self):
        n = len(self._values[~self._zero_mask])
        # Link above
        gamma = two_dim_vertical_first_deriv(1, n, factor=self._bounding_factor)
        bounded_altitudes = self._altitudes_m[~self._zero_mask] > self._upperbound
        first_idx = np.nonzero(bounded_altitudes)[0][0]
        bounded_altitudes[first_idx - 1] = True
        gamma[~bounded_altitudes, :] = 0
        self._inverse_Sa_bounding = gamma.T @ gamma

        # Link below
        gamma = two_dim_vertical_first_deriv(1, n, factor=self._bounding_factor)
        bounded_altitudes = self._altitudes_m[~self._zero_mask] < self._lowerbound
        gamma[~bounded_altitudes, :] = 0

        self._inverse_Sa_bounding += gamma.T @ gamma

        gamma = two_dim_vertical_second_deriv(
            1, n, factor=self._second_order_tikhonov_factor
        )
        self._inverse_Sa = gamma.T @ gamma

    def state(self) -> np.array:
        return copy(self._values[~self._zero_mask])

    def update_state(self, x: np.array):
        m_factors = np.exp(x - (self._values[~self._zero_mask]))
        m_factors[m_factors > self._max_update_factor] = self._max_update_factor
        m_factors[m_factors < 1 / self._max_update_factor] = 1 / self._max_update_factor
        self._values[~self._zero_mask] = np.log(
            copy(m_factors * np.exp(self._values[~self._zero_mask]))
        )
        self._internal_update()

    def apriori_state(self) -> np.array:
        x_a_bounding = self._initial_state
        full_inv_S_a = self._inverse_Sa + self._inverse_Sa_bounding
        return np.linalg.solve(full_inv_S_a, self._inverse_Sa_bounding @ x_a_bounding)

    def inverse_apriori_covariance(self) -> np.ndarray:
        return self._inverse_Sa_bounding + self._inverse_Sa

    def species(self):
        return self._species


class StateVectorProfileND(StateVectorProfile):
    def propagate_wf(self, radiance) -> xr.Dataset:
        interp_matrix = linear_interpolating_matrix(
            self._altitudes_m[self._state_mask], self._altitudes_m
        )
        wf_name = "wf_" + self.name()
        if wf_name not in radiance:
            msg = "WF not calculated"
            raise ValueError(msg)

        new_wf = radiance[wf_name]
        new_matrix = (
            self._values[:, np.newaxis]
            * interp_matrix
            / self._values[self._state_mask][np.newaxis, :]
        )
        return new_wf @ xr.DataArray(
            data=new_matrix,
            dims=["perturbation", "x"],
            coords={"perturbation": radiance["perturbation"].to_numpy()},
        )

    def __init__(
        self,
        altitudes_m: np.array,
        values: np.array,
        species_name: str,
        optical_property: sk.OpticalProperty,
        lowerbound: float,
        upperbound: float,
        second_order_tikhonov_factor: float,
        max_update_factor: float = 5,
    ):
        """

        Parameters
        ----------
        altitudes_m: np.array
            Altitudes of the profile, should match weighting function altitudes
        values: np.array
            log(number density) for the profile
        species_name: str
            Name of the species in the atmosphere
        optical_property: sk.OpticalProperty
            Optical property of the species
        lowerbound: float
            Lowerbound of the retrieval in m
        upperbound: float
            Upperbound of the retrieval in m
        second_order_tikhonov_factor: float
            Second order tikhonov factor
        max_update_factor: float, optional
            Maximum factor to update any element of the state vector every iteration. Default 5
        """
        self._altitudes_m = altitudes_m
        self._values = values
        self._lowerbound = lowerbound
        self._upperbound = upperbound

        self._state_mask = (self._altitudes_m > lowerbound) & (
            self._altitudes_m < upperbound
        )

        self._second_order_tikhonov_factor = second_order_tikhonov_factor
        self._max_update_factor = max_update_factor

        super().__init__(species_name, optical_property)

        self._initial_state = copy(self.state())
        self._compute_apriori_covariance()

    def _climatology_altitudes_and_density(self):
        values = copy(self._values)

        return self._altitudes_m, values

    def _compute_apriori_covariance(self):
        n = len(self._values[self._state_mask])
        gamma = two_dim_vertical_second_deriv(
            1, n, factor=self._second_order_tikhonov_factor
        )
        self._inverse_Sa = gamma.T @ gamma

    def state(self) -> np.array:
        return copy(self._values[self._state_mask])

    def update_state(self, x: np.array):
        m_factors = x / (self._values[self._state_mask])
        if self._max_update_factor is not None:
            m_factors[m_factors > self._max_update_factor] = self._max_update_factor
            m_factors[m_factors < 1 / self._max_update_factor] = (
                1 / self._max_update_factor
            )

        m_factors = np.interp(
            self._altitudes_m, self._altitudes_m[self._state_mask], m_factors
        )

        self._values = copy(m_factors * self._values)
        self._internal_update()

    def apriori_state(self) -> np.array:
        return self._initial_state

    def inverse_apriori_covariance(self) -> np.ndarray:
        return self._inverse_Sa

    def species(self):
        return self._species



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
import scipy.special as special
from numba import vectorize


class LineShape(ABC):
    """
    Base class for implementing line shapes.  A line shape represents integration across a high resolution measurement
    down to a lower resolution measurement.
    """

    @abstractmethod
    def integration_weights(
        self, mean: float, available_samples: np.ndarray, normalize=True
    ):
        """
        Integration weights for the line shape.

        Parameters
        ----------
        mean : float
            Value to integrate to

        available_samples : np.ndarray
            Array of sample values that are available to use in the integration.

        normalize : bool, Optional
            If true, resulting weights are normalized such that np.nansum(weights) = 1

        Returns
        -------
        np.ndarray
            Integration weights, same size as available_samples.
        """

    @abstractmethod
    def bounds(self, center=0):
        """
        Boundaries of the line shape.  Values outside this range are 0

        Parameters
        ----------
        center : float, optional
            Center of the line shape.  Default is 0

        Returns
        -------
        (left, right)
            Left and right boundaries of the line shape
        """

    def zero_centered(self):
        """
        True if the lineshape is centered on 0 rather than the nominal sample

        Returns
        -------
        bool
        """
        return True


class Gaussian(LineShape):
    def __init__(
        self,
        fwhm: float | None = None,
        stdev: float | None = None,
        max_stdev=5,
        mode="linear",
    ):
        """
        Gaussian line shape.

        Parameters
        ----------
        fwhm : float
            Full width half maximum, only specify one of fwhm or stdev

        stdev : float
            Standard deviation, only specify one of fwhm or stdev

        max_stdev: int, optional
            Values farther than max_stdev*stdev are truncated to 0.  Default=5

        mode : string, one of ['constant', 'linear']
             If constant, then the gaussian is sampled at the integration location.  If linear, then the gaussian
             is integrated with a triangular base function representing linear integration across the sample.  Linear
             is much more accurate at the cost of a small performance hit.  Default 'linear'.
        """

        self._fwhm_to_stdev = 1 / (2 * np.sqrt(2 * np.log(2)))

        if fwhm is not None and stdev is not None:
            msg = "Only one of fwhm or stdev should be specified"
            raise ValueError(msg)

        if fwhm is None and stdev is None:
            msg = "One of fwhm or stdev needs to be specified"
            raise ValueError(msg)

        if fwhm is not None:
            self._stdev = fwhm * self._fwhm_to_stdev
        else:
            self._stdev = stdev

        self.max_stdev = max_stdev

        self._mode = mode

    def integration_weights(
        self, mean: float, available_samples: np.ndarray, normalize=True
    ):
        """
        The lineshape converts a function at high resolution, H, to one at low resolution, L.

        L(mean) = np.dot(H, integration_weights(mean, available_samples))

        Parameters
        ----------
        mean : float
            value (in the range of available_samples) to integrate to
        available_samples : np.ndarray
            Values that the high resolution function is available at
        normalize : bool, optional
            If true, the returned weights will sum to 1

        Returns
        -------
        np.ndarray
            Weights to use in the high resolution integration
        """
        if self._stdev == 0:
            # Special case, just interpolate to the mean
            # TODO: This is nearest neighbor, switch to interpolate
            weights = np.zeros_like(available_samples)

            weights[np.argmin(np.abs(mean - available_samples))] = 1

            return weights

        difference = mean - available_samples

        if self._mode == "constant":
            gaussian = self._analytic_constant_weights(mean, available_samples)
        elif self._mode == "linear":
            gaussian = self._analytic_linear_weights(mean, available_samples)
        else:
            msg = "mode must be one of linear or constant"
            raise ValueError(msg)

        gaussian[np.abs(difference) > self.max_stdev * self._stdev] = 0
        # Sometimes numerical funniness can cause negative values
        # TODO: Check if this should be abs or set to 0
        gaussian = np.abs(gaussian)

        if normalize:
            if np.abs(np.sum(gaussian)) > 0:
                gaussian /= np.sum(gaussian)
            else:
                # Should be all zeros, indicating no contribution to the point which is okay
                # Just to be sure
                gaussian = np.zeros_like(gaussian)

        return gaussian

    def bounds(self, center=0):
        """
        If integration_weights is called with mean=0, all values outside the range [lower_bound, upper_bound] are
        guaranteed to be 0.

        Parameters
        ----------
        center : float, optional
            Center of the line shape.  Default is 0

        Returns
        -------
        [lower_bound, upper_bound]
        """
        return (
            -self.max_stdev * self._stdev - center,
            self.max_stdev * self._stdev - center,
        )

    def _analytic_linear_weights(self, mean, available_samples):
        return _gaussian_analytic_linear_weights(
            self.max_stdev, self._stdev, mean, np.ascontiguousarray(available_samples)
        )

    def _analytic_constant_weights(self, mean, available_samples):
        difference = mean - available_samples

        return np.exp(-0.5 * (difference / self._stdev) ** 2)


@vectorize("f8(f8)", nopython=True)
def fasterf1(x):
    """
    Fast error function approximation
    """
    p = 0.47047
    a1 = 0.3480242
    a2 = -0.0958798
    a3 = 0.7478556

    t = 1 / (1 + p * np.abs(x))

    return (
        1 - (a1 * t + a2 * t**2 + a3 * t**3) * np.exp(-np.abs(x) ** 2)
    ) * np.sign(x)


@vectorize("f8(f8)", nopython=True)
def fasterf2(x):
    """
    Fast error function approximation
    """
    p = 0.3275911
    a1 = 0.254829592
    a2 = -0.284496736
    a3 = 1.421413741
    a4 = -1.453152027
    a5 = 1.061405429

    t = 1 / (1 + p * np.abs(x))

    return (
        1
        - (a1 * t + a2 * t**2 + a3 * t**3 + a4 * t**4 + a5 * t**5)
        * np.exp(-np.abs(x) ** 2)
    ) * np.sign(x)


def fasterf3(x):
    """
    Fast error function approximation
    """
    return special.erf(x)


@vectorize("f8(f8, f8, f8, f8)")
def _gaussian_analytic_linear_weights_helper(width_left, width_right, offsets, stdev):
    """
    See theory/line_shape_integrals.nb for explanation
    """
    stdev /= np.sqrt(0.5)

    return (
        0.5
        * stdev
        * (
            stdev
            * (
                np.exp(-((offsets - width_left) ** 2) / stdev**2) / width_left
                + np.exp(-((offsets + width_right) ** 2) / stdev**2) / width_right
                - np.exp(-(offsets**2) / stdev**2)
                * (width_left + width_right)
                / (width_left * width_right)
            )
            + 1
            / (width_left * width_right)
            * np.sqrt(np.pi)
            * (
                -offsets * (width_right + width_left) * fasterf1(offsets / stdev)
                + (offsets - width_left)
                * width_right
                * fasterf1((offsets - width_left) / stdev)
                + width_left
                * (offsets + width_right)
                * fasterf1((offsets + width_right) / stdev)
            )
        )
    )


def _gaussian_analytic_linear_weights(max_stdev, stdev, mean, available_samples):
    """
    Calculates the weights by integrating the gaussian form with a triangle function
    """
    weights = np.zeros_like(available_samples)

    # Difference between center of the gaussian and the sample
    offsets = available_samples - mean
    within_stdev = np.abs(offsets) < max_stdev * stdev

    # Interpolation width on the left side of each sample
    widths = np.diff(available_samples)
    width_left = np.abs(np.hstack((np.array([widths[0]]), widths)))

    # Interpolation width on the right side of each sample
    width_right = np.abs(np.hstack((widths, np.array([widths[-1]]))))

    offsets = offsets[within_stdev]
    width_left = width_left[within_stdev]
    width_right = width_right[within_stdev]

    # Formula derived by integrating a triangle function with a gaussian
    weights[within_stdev] = _gaussian_analytic_linear_weights_helper(
        width_left, width_right, offsets, stdev
    )

    return weights


class DeltaFunction(LineShape):
    def __init__(self):
        """
        DeltaFunction line shape.  The nearest sample is always taken.

        """

    def integration_weights(
        self,
        mean: float,
        available_samples: np.ndarray,
        normalize=True,  # noqa: ARG002
        tolerance=1e-7,
    ):
        # Interpolate to the mean value
        weights = np.zeros_like(available_samples)

        if np.sum(np.abs(mean - available_samples) < tolerance) > 1:
            weights[np.where(np.abs(mean - available_samples) < tolerance)[0]] = 1
        else:
            weights[np.argmin(np.abs(mean - available_samples))] = 1

        return weights

    def bounds(self, center=0):
        return center, center


class Rectangle(LineShape):
    def __init__(self, width, mode="linear"):
        """
        Rectangular line shape

        Parameters
        ----------
        width : float
            Full width of the line shape.
        """
        self._width = width
        self._mode = mode

    def integration_weights(
        self, mean: float, available_samples: np.ndarray, normalize=True
    ):
        if self._mode == "linear":
            weights = self._analytic_linear_weights(mean, available_samples)
        else:
            weights = np.zeros_like(available_samples)

            offsets = mean - available_samples

            weights[np.abs(offsets) < self._width / 2] = 1

        if normalize:
            if np.nansum(weights) == 0:
                raise ValueError()
            weights /= np.nansum(weights)

        return weights

    def _analytic_linear_weights(self, mean, available_samples):
        """
        See theory/line_shape_integrals.nb for more information.  Linear weights are calculated by integrating
        the rectangle function with a triangle function.
        """
        weights = np.zeros_like(available_samples)

        # Difference between center of the gaussian and the sample
        offsets = mean - available_samples
        # within_width = np.abs(offsets) < self._width

        # Interpolation width on the left side of each sample
        widths = np.diff(available_samples)

        width_left = np.abs(np.hstack((np.array([widths[0]]), widths)))

        # Interpolation width on the right side of each sample
        width_right = np.abs(np.hstack((widths, np.array([widths[-1]]))))

        # offsets = offsets[within_width]
        # width_left = width_left[within_width]
        # width_right = width_right[within_width]

        weights += _rectangle_analytic_linear_weights_helper_left(
            width_left, offsets, self._width
        )
        weights += _rectangle_analytic_linear_weights_helper_right(
            width_right, offsets, self._width
        )

        return weights

    def bounds(self, center=0):
        return [-self._width / 2 + center, self._width / 2 + center]


@vectorize("f8(f8, f8, f8)", nopython=True)
def _rectangle_analytic_linear_weights_helper_left(width_left, offset, rect_width):
    if offset == 0 and rect_width < 2 * width_left:
        return -rect_width * (rect_width - 4 * width_left) / (8 * width_left)
    elif (  # noqa: RET505
        offset > 0
        and 2 * offset + rect_width < 2 * width_left
        and 2 * offset < rect_width
    ) or (
        offset < 0
        and 2 * offset + rect_width > 0
        and 2 * offset + rect_width < 2 * width_left
    ):
        return (
            -(2 * offset + rect_width)
            * (2 * offset + rect_width - 4 * width_left)
            / (8 * width_left)
        )
    elif (
        offset > 0
        and 2 * offset + rect_width <= 2 * width_left
        and (2 * offset == rect_width or (rect_width > 0 and 2 * offset >= rect_width))
    ):
        return rect_width - offset * rect_width / width_left
    elif width_left > 0 and (
        (offset == 0 and rect_width > 2 * width_left)
        or (2 * offset + rect_width >= 2 * width_left and offset < 0)
        or (
            offset > 0
            and 2 * offset + rect_width > 2 * width_left
            and 2 * offset < rect_width
        )
    ):
        return width_left / 2
    elif (
        offset > 0
        and 2 * offset + rect_width > 2 * width_left
        and (
            (2 * offset == rect_width and width_left < 0)
            or (2 * offset > rect_width and 2 * offset < rect_width + 2 * width_left)
        )
    ):
        return (-2 * offset + rect_width + 2 * width_left) ** 2 / (8 * width_left)
    else:
        return 0


@vectorize("f8(f8, f8, f8)", nopython=True)
def _rectangle_analytic_linear_weights_helper_right(width_right, offset, rect_width):
    if width_right > 0 and (
        (offset == 0 and rect_width > 0 and rect_width > 2 * width_right)
        or (
            2 * (offset + width_right) < rect_width
            and (
                (2 * offset + rect_width > 0 and offset < 0)
                or (offset > 0 and 2 * offset < rect_width)
            )
        )
    ):
        return width_right / 2
    elif offset < 0 and (  # noqa: RET505
        (2 * offset + rect_width == 0 and 2 * (offset + width_right) >= rect_width)
        or (
            rect_width > 0
            and 2 * (offset + width_right) > rect_width
            and 2 * offset + rect_width < 0
        )
    ):
        return rect_width * (offset + width_right) / width_right
    elif offset == 0 and rect_width > 0 and rect_width <= 2 * width_right:
        return -rect_width * (rect_width - 4 * width_right) / (8 * width_right)
    elif 2 * (offset + width_right) >= rect_width and (
        (2 * offset + rect_width > 0 and offset < 0)
        or (offset > 0 and 2 * offset < rect_width)
    ):
        return (
            -(2 * offset - rect_width)
            * (2 * offset - rect_width + 4 * width_right)
            / (8 * width_right)
        )
    elif (
        2 * offset + rect_width + 2 * width_right > 0
        and offset < 0
        and (
            (2 * offset + rect_width == 0 and 2 * (offset + width_right) < rect_width)
            or (
                rect_width > 0
                and 2 * offset + rect_width < 0
                and 2 * (offset + width_right) <= rect_width
            )
        )
    ):
        return (2 * offset + rect_width + 2 * width_right) ** 2 / (8 * width_right)
    else:
        return 0


class UserLineShape(LineShape):
    def __init__(
        self,
        x_values: np.array,
        line_values: np.array,
        zero_centered: bool,
        integration_fraction: float = 0.999,
        mode="simple",
    ):
        """
        Line shape created from a user specified function

        Parameters
        ----------
        x_values: np.array
            x values for the lineshape, could be wavelength, could be angle, etc.
        line_values: np.array
            Values for the line shape.  Same size as x_values.  Any values outside the range of x_values are assumed
            to be 0.
        zero_centered: bool
            True if the line shape values are centered at 0, false if the line shape is not centered.
        integration_fraction: float
            Fraction of the line shape to integrate.  Default is 1, which means the entire line shape is integrated.
        mode: str
            If set to 'simple', the line shape is interpolated to the sample values.  If mode is set to 'integrate'
            then the line shape is analytically integrated assuming linear interpolation over the sample values.
            If the mode is 'integrate' then the line shape samples must be evenly spaced.

        """
        self._x_values = x_values
        self._line_values = line_values
        self._zero_centered = zero_centered

        center = np.argmax(np.abs(self._line_values))
        right_integral = np.cumsum(np.abs(self._line_values[center:]))
        left_integral = np.cumsum(np.abs(self._line_values)[:center][::-1])

        try:
            self._right_cutoff = (
                np.nonzero(right_integral / right_integral[-1] >= integration_fraction)[
                    0
                ][0]
                + center
            )
        except IndexError:
            self._right_cutoff = len(self._line_values) - 1
        try:
            self._left_cutoff = (
                center
                - np.nonzero(left_integral / left_integral[-1] >= integration_fraction)[
                    0
                ][0]
            )
        except IndexError:
            self._left_cutoff = 0

        self._mode = mode
        if self._mode == "integrate":
            self._widths_left = np.zeros_like(self._x_values, dtype=np.float64)
            self._widths_right = np.zeros_like(self._x_values, dtype=np.float64)

            self._widths_left[1:] = self._x_values[1:] - self._x_values[:-1]
            self._widths_left[0] = 1e99
            self._widths_right[:-1] = self._x_values[1:] - self._x_values[:-1]
            self._widths_right[-1] = 1e99

    def integration_weights(
        self, mean: float, available_samples: np.ndarray, normalize=True
    ):
        # interpolate the line shape to the available samples

        x_interp = self._x_values + mean if self._zero_centered else self._x_values

        if self._mode == "simple":
            line_shape_interp = np.interp(
                available_samples, x_interp, self._line_values, left=0, right=0
            )
        elif self._mode == "integrate":
            line_shape_interp = self._linear_weights(mean, available_samples, x_interp)
        else:
            msg = "UserLineShape mode must be one of simple or integrate"
            raise ValueError(msg)

        if not normalize:
            msg = "UserLineShape currently only supports normalized line shapes"
            raise ValueError(msg)

        line_shape_interp /= np.sum(line_shape_interp)

        return line_shape_interp

    def bounds(self, center=0):
        return (
            self._x_values[self._left_cutoff] + center,
            self._x_values[self._right_cutoff] + center,
        )
        return np.min(self._x_values) + center, np.max(self._x_values) + center

    def zero_centered(self):
        return self._zero_centered

    def _linear_weights(self, mean, available_samples, xs):
        offsets = mean - available_samples if self._zero_centered else available_samples

        # Interpolation width on the left side of each sample
        widths = np.diff(available_samples)

        width_left = np.abs(np.hstack((np.array([widths[0]]), widths)))

        # Interpolation width on the right side of each sample
        width_right = np.abs(np.hstack((widths, np.array([widths[-1]]))))

        # Have to sum over all internal x values
        weights = np.zeros_like(available_samples)
        temp = np.zeros_like(available_samples)
        for x, ls, wl, wr in zip(
            xs, self._line_values, self._widths_left, self._widths_right
        ):
            weights += (
                _triangle_analytic_linear_weights_helper2(
                    wr, -1 * (offsets - x), width_right
                )
                * ls
            )
            # Integral should be symmetric around x->-x and offset->-offset so use same helper for both cases
            weights += (
                _triangle_analytic_linear_weights_helper2(wl, offsets - x, width_left)
                * ls
            )

            temp += (
                _triangle_analytic_linear_weights_helper2(
                    wr, -1 * (offsets - x), width_right
                )
                * ls
            )
            temp += (
                _triangle_analytic_linear_weights_helper2(wl, offsets - x, width_left)
                * ls
            )

        return weights


@vectorize("f8(f8, f8, f8)", nopython=True)
def _triangle_analytic_linear_weights_helper(width_right, offset, width):
    if offset < 0 and (offset + width_right) > width:
        return (
            -1.0
            * (width * (-3.0 * offset + width - 3.0 * width_right))
            / (6.0 * width_right)
        )
    elif (  # noqa: RET505
        offset < 0 and (offset + width_right) > 0 and (offset + width_right) <= width
    ):
        return (
            -1.0
            * (offset + width_right) ** 2
            * (offset - 3.0 * width + width_right)
            / (6.0 * width * width_right)
        )
    elif (offset + width_right) > width and (
        offset == 0 or ((width > offset) and offset > 0)
    ):
        return (
            (offset - width) ** 2
            * (offset - width + 3.0 * width_right)
            / (6.0 * width * width_right)
        )
    elif (offset + width_right) <= width and offset >= 0:
        return (
            -1.0
            * width_right
            * (3.0 * offset - 3.0 * width + width_right)
            / (6.0 * width)
        )
    else:
        return 0.0


@vectorize("f8(f8, f8, f8)", nopython=True)
def _triangle_analytic_linear_weights_helper2(width_right, offset, width):
    if (
        offset < 0
        and offset + width_right <= width
        and (
            (offset + width == 0 and offset + width_right < 0)
            or (
                offset + width < 0
                and (
                    offset + width_right == 0
                    and (offset + width + width_right > 0 and offset + width_right <= 0)
                )
            )
        )
    ):
        return (offset + width + width_right) ** 3 / (6 * width * width_right)
    elif offset + width_right < 0 and offset + width > 0:  # noqa: RET505
        return (width_right * (3 * (offset + width) + width_right)) / (6 * width)
    elif offset > 0 and offset < width and offset + width_right > width:
        return ((offset - width) ** 2 * (offset - width + 3 * width_right)) / (
            6 * width * width_right
        )
    elif offset + width_right > width and offset + width <= 0:
        return (width * (offset + width_right)) / width_right
    elif offset == 0 and width < width_right:
        return -1 * (width * (width - 3 * width_right)) / (6 * width_right)
    elif offset + width == 0 and offset + width_right == 0:
        return width_right / 6
    elif offset + width_right == 0 and offset + width > 0:
        return ((3 * width - 2 * width_right) * width_right) / (6 * width)
    elif offset < 0 and offset + width > 0 and offset + width_right > width:
        return (
            -1
            * (
                offset**3
                + width**2 * (width - 3 * width_right)
                - 3 * offset * width * (width - 2 * width_right)
                + 3 * offset**2 * (width + width_right)
            )
            / (6 * width * width_right)
        )
    elif offset == 0 and width >= width_right:
        return ((3 * width - width_right) * width_right) / (6 * width)
    elif offset > 0 and offset + width_right <= width:
        return -1 * (width_right * (3 * offset - 3 * width + width_right)) / (6 * width)
    elif (
        offset < 0
        and offset + width_right > 0
        and offset + width > 0
        and offset + width_right <= width
    ):
        return (
            -1
            * (
                2 * offset**3
                + 6 * offset**2 * width_right
                + 3 * (offset - width) * width_right**2
                + width_right**3
            )
            / (6 * width * width_right)
        )
    elif (
        offset + width <= 0
        and offset + width_right > 0
        and offset + width_right <= width
    ):
        return (
            -(offset**3)
            + width**3
            + 3 * offset**2 * (width - width_right)
            + 3 * width**2 * width_right
            + 3 * width * width_right**2
            - width_right**3
            + 3 * offset * (width**2 + 2 * width * width_right - width_right**2)
        ) / (6 * width * width_right)
    else:
        return 0.0


def _triangle_analytic_linear_weights_helper3(width_right, offset, width):
    if (
        offset < 0
        and offset + width_right <= width
        and (
            (offset + width == 0 and offset + width_right < 0)
            or (
                offset + width < 0
                and (
                    offset + width_right == 0
                    and (offset + width + width_right > 0 and offset + width_right <= 0)
                )
            )
        )
    ):
        return (offset + width + width_right) ** 3 / (6 * width * width_right)
    elif offset + width_right < 0 and offset + width > 0:  # noqa: RET505
        return (width_right * (3 * (offset + width) + width_right)) / (6 * width)
    elif offset > 0 and offset < width and offset + width_right > width:
        return ((offset - width) ** 2 * (offset - width + 3 * width_right)) / (
            6 * width * width_right
        )
    elif offset + width_right > width and offset + width <= 0:
        return (width * (offset + width_right)) / width_right
    elif offset == 0 and width < width_right:
        return -1 * (width * (width - 3 * width_right)) / (6 * width_right)
    elif offset + width == 0 and offset + width_right == 0:
        return width_right / 6
    elif offset + width_right == 0 and offset + width > 0:
        return ((3 * width - 2 * width_right) * width_right) / (6 * width)
    elif offset < 0 and offset + width > 0 and offset + width_right > width:
        return (
            -1
            * (
                offset**3
                + width**2 * (width - 3 * width_right)
                - 3 * offset * width * (width - 2 * width_right)
                + 3 * offset**2 * (width + width_right)
            )
            / (6 * width * width_right)
        )
    elif offset == 0 and width >= width_right:
        return ((3 * width - width_right) * width_right) / (6 * width)
    elif offset > 0 and offset + width_right <= width:
        return -1 * (width_right * (3 * offset - 3 * width + width_right)) / (6 * width)
    elif (
        offset < 0
        and offset + width_right > 0
        and offset + width > 0
        and offset + width_right <= width
    ):
        return (
            -1
            * (
                2 * offset**3
                + 6 * offset**2 * width_right
                + 3 * (offset - width) * width_right**2
                + width_right**3
            )
            / (6 * width * width_right)
        )
    elif (
        offset + width <= 0
        and offset + width_right > 0
        and offset + width_right <= width
    ):
        return (
            -(offset**3)
            + width**3
            + 3 * offset**2 * (width - width_right)
            + 3 * width**2 * width_right
            + 3 * width * width_right**2
            - width_right**3
            + 3 * offset * (width**2 + 2 * width * width_right - width_right**2)
        ) / (6 * width * width_right)
    else:
        return 0.0
